# -*- python -*-
# $RCSfile: fileselector.py,v $
# $Revision: 1.22 $
# $Author: langer $
# $Date: 2006/03/09 20:22:20 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from oof2.SWIG.common import switchboard
from oof2.common import debug
from oof2.common.IO import parameter
from oof2.common.IO import reporter
from oof2.common.IO.GUI import oof_mainiteration
from oof2.common.IO.GUI import parameterwidgets
from oof2.common.IO.GUI import widgetscope
import gtk
import os
import weakref

# The _selectors dictionary keeps track of file selectors to be
# reused.  They are reused so that they preserve their state (current
# directory, etc).  The dictionary is a WeakValueDictionary because
# the selectors can be destroyed by the window manager, and we don't
# have any control over that.

_selectors = weakref.WeakValueDictionary()


modes = {'w':gtk.FILE_CHOOSER_ACTION_SAVE,
         'r':gtk.FILE_CHOOSER_ACTION_OPEN}

class FileSelector(widgetscope.WidgetScope):
    OK = 1
    CANCEL = 2
    def __init__(self, mode, title=None, filename=None, params=None):
        debug.mainthreadTest()
        widgetscope.WidgetScope.__init__(self, None)
        self.dialog = gtk.FileChooserDialog(
            title=title or "OOF2 File Selector",
            action=modes[mode],
            buttons=(gtk.STOCK_OK, self.OK,
                     gtk.STOCK_CANCEL, self.CANCEL)
            )
        self.dialog.set_default_response(self.OK)
        if params is None:
            self.table = None
        else:
            self.table = parameterwidgets.ParameterTable(params, scope=self)
            self.dialog.set_extra_widget(self.table.gtk)

        if filename is not None:
            self.dialog.set_current_name(filename)

        self.dialog.connect('selection-changed', self.selectionchangedCB)
        self.sbcallback = switchboard.requestCallback(('validity', self.table),
                                                      self.validityCB)
        self.sensitize()
    def set_mode(self, mode):
        self.dialog.set_action(modes[mode])
        self.sensitize()
    def sensitize(self):
        valid = self.table is None or self.table.isValid()
        # If we're opening an existing file, check to see if it's
        # really there.  If we're opening a new file, it would be nice
        # to check to see if the file entry widget has text in it, but
        # there seems to be no way to do that.  There's no way to call
        # this function after every keystroke, since the Entry widget
        # inside the FileChooserDialog isn't accessible and we can't
        # connect to it.
        if valid and self.dialog.get_action() == gtk.FILE_CHOOSER_ACTION_OPEN:
            filename = self.dialog.get_filename()
            valid = valid and filename is not None and os.path.isfile(filename)
        self.dialog.set_response_sensitive(self.OK, valid)
    def validityCB(self, validity):
        self.sensitize()
    def selectionchangedCB(self, *args):
        self.sensitize()
    def run(self):
        debug.mainthreadTest()
        return self.dialog.run()
    def get_values(self):
        if self.table:
            self.table.get_values()
    def getFileName(self):
        if self.table:
            self.table.get_values()
        return self.dialog.get_filename()
    def close(self):
        debug.mainthreadTest()
        switchboard.removeCallback(self.sbcallback)
        self.dialog.destroy()
        self.destroyScope()
    def hide(self):
        debug.mainthreadTest()
        self.dialog.hide()
        
# Always check the return value of fileSelector()!  fileSelector
# returns None if the user cancels the operation or doesn't type in a
# name in 'w' mode.

# mode should be 'r' or 'w'.

def fileSelector(mode, ident=None, title=None, filename=None, params=None):
    if _selectors.has_key(ident):
        selector = _selectors[ident]
        selector.set_mode(mode)
    else:
        selector = FileSelector(mode=mode, title=title, filename=filename,
                                params=params)
        if ident is not None:
            _selectors[ident] = selector

    result = selector.run()
    selector.hide()

    if result in (FileSelector.CANCEL,
                  gtk.RESPONSE_DELETE_EVENT,
                  gtk.RESPONSE_NONE):
        return None
    return selector.getFileName()

# This function should be called before you write to a file.
# If the file exists, it will prompt the user, asking whether to
# over-write or append.  It will return either 'w' for writing,
# 'a' for appending, or None, in which case the caller should not
# write to the file at all.
def get_file_mode(filename, no_append=0):
    mode = None
    try:
        os.stat(filename)               # does file exist?
    except OSError:                     # file not found
        mode = 'w'
    else:
        if no_append:
            ans = reporter.query("Overwrite file\n%s?" % filename,
                                 "OK", "Cancel", default="OK")
            if ans == "OK":
                mode = 'w'
        else:
            ans = reporter.query("Overwrite file\n%s?" % filename,
                                 "OK", "Append", "Cancel", default="OK")
            if ans == 'OK':
                mode = 'w'
            elif ans == 'Append':
                mode = 'a'
    return mode
    
