// -*- C++ -*-
// $RCSfile: cmicrostructure.h,v $
// $Revision: 1.34 $
// $Author: reida $
// $Date: 2006/03/15 22:18:02 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#ifndef CMICROSTRUCTURE_H
#define CMICROSTRUCTURE_H

#include <Python.h>
#include <vector>
#include "common/array.h"
#include "common/boolarray.h"
#include "common/coord.h"
#include "common/pixelgroup.h"
#include "common/timestamp.h"
#include "common/pixelsetboundary.h"


class ActiveArea;
class CSegment;
class CRectangle;
class PixelAttribute;

// Some operations, such as finding the pixels under an element,
// require marking pixels in the microstructure.  Neither the
// microstructure nor the pixels can keep track of which pixels are
// marked, because there might be concurrent marking threads.  So the
// mark information is kept in a separate MarkInfo object.
class MarkInfo {
private:
  friend class CMicrostructure;
  MarkInfo(const ICoord &size);
  BoolArray markedpixels;
  BoolArray markedregion;	// active subarray of markedpixels
  void mark_site(std::vector<ICoord>&, const ICoord&);
  friend std::ostream &operator<<(std::ostream &os, const MarkInfo &mi) {
    os << mi.markedregion; 
    return os;
  }
};


class CMicrostructure {
private:
  ICoord pxlsize_;		// size of microstructure in pixels
  Coord size_;			// physical size of whole microstructure
  Coord delta_;			// physical size of a pixel
  TimeStamp timestamp;

  // List of pixel groups defined on this microstructure.
  std::vector<PixelGroup*> pixelgroups;
  // When groups are removed from the microstructure, they're not
  // deleted right away, because it would take too long to remove them
  // from the attributeMap.  They're put in the defunctgroups list until
  // the next categorization pass.
  mutable std::vector<PixelGroup*> defunctgroups;

  // List of arrays of PixelAttributes.  PixelAttributes allow other
  // modules to assign properties to pixels.  categorize() assigns
  // different categories to pixels with different properties, and the
  // mesh generator attempts to segregate different categories into
  // separate elements.  See comments in pixelattribute.h
  mutable std::vector<Array<PixelAttribute*> > attributeMap;
  std::vector<PixelAttributeGlobalData*> attributeGlobalData;

  // categorymap caches the pixel categories assigned by categorize().
  // It's mutable because CMicrostructure::category() is
  // logically const, but it caches the categories in the categorymap.
  mutable Array<int> categorymap;
  // representativePixels is sort of the inverse of categorymap. Given
  // a category, it tells you one pixel in that category, from which
  // you can determine the actual PixelAttributes of the category.
  mutable std::vector<ICoord> representativePixels;

  // List of the boundaries of the categories.
  mutable std::vector<PixelSetBoundary> categoryBdys;

  mutable bool categorized;
  mutable int ncategories;
  void categorize() const;

  ActiveArea *activearea;

  std::string name_;
public:
  CMicrostructure(const std::string &name,
		  const ICoord *pxlsize, const Coord *size);
  ~CMicrostructure();
  const std::string &name() const { return name_; }
  void rename(const std::string &name) { name_ = name; }
  void destroy();
  const Coord &size() const { return size_; }
  const ICoord &sizeInPixels() const { return pxlsize_; }
  const Coord &sizeOfPixels() const { return delta_; }
  Coord physical2Pixel(const Coord&) const; // real space to pixel coords
  Coord pixel2Physical(const ICoord&) const;
  Coord pixel2Physical(const Coord&) const;
  ICoord pixelFromPoint(const Coord&) const; // pixel containing the given point
  bool contains(const ICoord&) const;
  TimeStamp getTimeStamp() const;

  void setCurrentActiveArea(ActiveArea *aa) { activearea = aa; }
  const ActiveArea *getActiveArea() const { return activearea; }

  int nGroups() const;
  PixelGroup *getGroup(const std::string &name, bool *newness);
  PixelGroup *findGroup(const std::string &name) const;
  void removeGroup(const std::string &name);
  std::vector<std::string> *groupNames() const;

  Array<PixelAttribute*> &getAttributeMap(int attributeID) const;
  PixelAttributeGlobalData *getAttributeGlobalData(int attributeID) const;
  const Array<int> *getCategoryMap() const; // changes mutable private data
  const Array<int> *getCategoryMapRO() const; // changes no data

  int nCategories() const;
  // Three different versions of this for convenience in calling it...
  int category(const ICoord *where) const;
  int category(const ICoord &where) const;
  int category(int x, int y) const;
  int category(const Coord &where) const; // Arbitrary phyiscal-coord point.
  void recategorize();
  const ICoord &getRepresentativePixel(int category) const;
  bool is_categorized() const { return categorized; }

  const std::vector<PixelSetBoundary> &getCategoryBdys() {
    return categoryBdys; }

  std::vector<ICoord> *segmentPixels(const CSegment*) const;
  std::vector<ICoord> *segmentPixels(const Coord&, const Coord&) const;
  
  MarkInfo *beginMarking(const CRectangle&) const; // sets active subarray of markedpixels
  void markSegment(MarkInfo*, const Coord&, const Coord&) const;
  void markTriangle(MarkInfo*, const Coord&, const Coord&, const Coord&)
    const;
  std::vector<ICoord> *markedPixels(MarkInfo*) const;
  void endMarking(MarkInfo*) const;

  bool transitionPoint(const Coord&, const Coord&, Coord *result) const;
  double edgeHomogeneity(const Coord&, const Coord&) const;
}; 




#endif // CMICROSTRUCTURE_H
