# -*- python -*-
# $RCSfile: analyze.py,v $
# $Revision: 1.47 $
# $Author: langer $
# $Date: 2006/01/30 22:18:56 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Objects for performing operations on data.  These include writing
# data to a file, taking averages and standard deviations, and so
# forth.  The data to which this should be done is specified by means
# of an output.

from oof2.common import debug
from oof2.common import enum
from oof2.common import primitives
from oof2.common import registeredclass
from oof2.common.IO import destinationparam
from oof2.common.IO import parameter
from oof2.common.IO import whoville
from oof2.common.IO import xmlmenudump
from oof2.SWIG.common import ooferror
from oof2.SWIG.engine import outputval
import math
import string
import types


# Parent class for all "Data Operation" objects, i.e. things which do
# things to outputs, like writing them gridwise to files,
# evaluating them along cross-sections, and so forth.

# Data operation objects need to specify at registration time
# whether they are "direct" or not -- a "direct" operation is one
# for which the value of the output at a given sample is directly
# output to the user.  Statistical operations are not direct.


# Utililty function, returns the output type of the output object, if
# it's one we can handle, or raises an exception.
def _ocheck(output):
    try:
        output.otype(types.FloatType)
        return types.FloatType
    except TypeError:
        try:
            output.otype(outputval.OutputValPtr)
            return outputval.OutputValPtr
        except TypeError:
            raise ooferror.ErrSetupError(
                "Unable to process %s from output %s." % \
                (`output.otype`,`output`) )



class DataOperation(registeredclass.RegisteredClass):
    registry = []
    tip='Post-processing data operations'


class Separator(enum.EnumClass("space", "comma", "tab")):
    tip="Characters used between columns in output files."
    discussion="""<para>
    The commands in the <xref linkend='MenuItem:OOF.Mesh.Analyze'/>
    menu write columns of data to output files.  The commands have a
    <varname>separator</varname> argument that sets the character that
    divides one column of data from the next.  The
    <classname>Separator</classname> contains all of the allowed
    separators.
    </para>"""

separator_strings = {
    Separator("space") : " ",
    Separator("comma") : ", ",
    Separator("tab") : "\t"}


# The most basic, a direct-output class that just writes out the
# data at all the samples. 
class DirectOutput(DataOperation):
    def __call__(self, output, domain, sampling, destination,
                 comment_character, separator):

        otype = _ocheck(output)
        
        header = sampling.make_samples(domain)
        sep_string = separator_strings[separator]

        # "olist" is a list of tuples of the form (sample, value).
        olist = sampling.evaluate(domain, output, otype)
        if otype == types.FloatType:
            output_header = header + ["Value"]
        else:
            # Pull the label list from the first value of the first
            # tuple in the list.
            output_header = header + olist[0][1].label_list()

        print >> destination, comment_character + \
              string.join(output_header, sep_string)

        if otype == types.FloatType:
            # For floating-point values, iterate over the list, and
            # just use the repr of the (possibly multiple) values.
            for (s,v) in olist:
                tags = s.columnData(header)
                if len(tags)==1:
                    print >> destination, \
                          string.join(tags[0]+[`v`], sep_string)
                else:
                    for (tag, vstring) in zip(tags, [`val` for val in v]):
                        print >> destination, \
                              string.join(tag+[vstring], sep_string)
        else: # otype == outputval.OutputValPtr
            # For the OutputVal case, use value_list to get the
            # list of floats, and join them up.
            for (s,v) in olist:
                tags = s.columnData(header)
                if len(tags)==1:
                    vlist = v.value_list()
                    print >> destination, \
                          string.join(tags[0] + [`x` for x in vlist],
                                      sep_string)
                else: # Multiple values -- do above for each tag-val pair.
                    for (tag, val) in zip(tags, v):
                        vlist = val.value_list()
                        print >> destination, \
                              string.join(tag + [`x` for x in vlist],
                                          sep_string)

# "Direct Output" is special, in that the corresponding auto-generated
# menu item is used directly in the meshcstoolboxGUI code -- if the
# name is changed here, it should be changed there also.                 
registeredclass.Registration(
    'Direct Output', DataOperation, DirectOutput, 1,
    params=[],
    tip="Write the data values directly.",
    direct=1,
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/directoutput.xml'))


# There are constraints on the output types of the output objects that
# can be processed by these functions -- they must either be Floats,
# or they must be composite objects for which __add__ and __sub__
# operate component-wise, __mul__ and __div__ take scalar operands,
# and for which point-wise operations "component_square" and
# "component_sqrt" are defined.  This is currently true of all
# OutputValPtr subclasses, including SymmMatrix3.


# Integrates the output over each sample, and normalizes by
# the power=0 version.
class AverageOutput(DataOperation):
    def __call__(self, output, domain, sampling, destination,
                 comment_character, separator):

        otype = _ocheck(output)

        if not sampling.fill_sample_list(domain):
            return                      # no samples!
        integralsum = None
        areasum = 0.0

        # The following line relies on the fact that integrate does
        # not evaluate its 'output' arg when power==0 and
        # otype==FloatType.
        for v in sampling.integrate(domain, None, types.FloatType, power=0):
            areasum += v[1]

        for v in sampling.integrate(domain, output, otype):
            if integralsum is not None:
                integralsum += v[1]
            else:
                integralsum = v[1]

        result = integralsum/areasum

        if otype==types.FloatType:
            print >> destination, comment_character + " Value"
            print >> destination, `result`
        else:
            sep_string = separator_strings[separator]
            output_header = result.label_list()
            print >> destination, comment_character + \
                  string.join(output_header, sep_string)
            print >> destination, \
                  string.join([`x` for x in result.value_list()],
                              sep_string)
                  

registeredclass.Registration(
    "Average",
    DataOperation,
    AverageOutput,
    2,
    direct=0,
    tip="Average the data over all the samples.",
    discussion="""<para>

    Compute the average value of the data over all of the samples.
    When integrating over element areas, the average is weighted by
    area.  Otherwise each sample gets equal weight.

    </para>"""
    )



# Standard deviation, square root of the integral of ( (x-<x>)^2 ),
# which is of course (<x^2>-<x>^2)^(1/2)
class StdDevOutput(DataOperation):
    def __call__(self, output, domain, sampling, destination,
                 comment_character, separator):

        otype = _ocheck(output)

        if not sampling.fill_sample_list(domain):
            return
        integralsum = None
        areasum = 0.0
        squaresum = None
        # The following line relies on the fact that integrate does
        # not evaluate its 'output' arg when power==0 and
        # otype==FloatType.
        for v in sampling.integrate(domain, None, types.FloatType, power=0):
            areasum += v[1]
            
        for v in sampling.integrate(domain, output, otype, power=1):
            if integralsum is None:
                integralsum = v[1]
            else:
                integralsum += v[1]

        for v in sampling.integrate(domain, output, otype, power=2):
            if squaresum is None:
                squaresum = v[1]
            else:
                squaresum += v[1]


        # Need different kinds of square roots depending on the
        # output type.
        if otype == types.FloatType:
            result = math.sqrt( (squaresum/areasum) -
                                (integralsum/areasum)**2 )
            print >> destination, comment_character + " Value"
            print >> destination, `result`

        elif otype == outputval.OutputValPtr:
            avgsquared = (integralsum/areasum)
            avgsquared.component_square()
            squaredavg = (squaresum/areasum)
            result = (squaredavg-avgsquared)
            result.component_sqrt()

            sep_string = separator_strings[separator]
            output_header = result.label_list()
            print >> destination, comment_character + \
                  string.join(output_header, sep_string)
            print >> destination, \
                  string.join([`x` for x in result.value_list()],
                              sep_string)

registeredclass.Registration(
    "Standard Deviation",
    DataOperation,
    StdDevOutput,
    3,
    direct=0,
    tip="Compute the standard deviation of the data over the samples.",
    discussion="""<para>
    Compute the standard deviation of the data over all of the samples.
    When integrating over element areas, the deviation is weighted by
    area.  Otherwise each sample gets equal weight.
    </para>""")

