# -*- python -*-
# $RCSfile: displaymethods.py,v $
# $Revision: 1.94 $
# $Author: langer $
# $Date: 2006/02/17 20:00:34 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Simple display methods for Skeletons and Meshes.  Simpler than
# contour plots, in any case.

from oof2.SWIG.common import devel
from oof2.SWIG.common import lock
from oof2.SWIG.common import ooferror
from oof2.SWIG.common import switchboard
from oof2.common import color
from oof2.common import debug
from oof2.common import mainthread
from oof2.common import primitives
from oof2.common import utils
from oof2.common.IO import automatic
from oof2.common.IO import colormap
from oof2.common.IO import display
from oof2.common.IO import ghostgfxwindow
from oof2.common.IO import oofmenu
from oof2.common.IO import output
from oof2.common.IO import parameter
from oof2.common.IO import reporter
from oof2.common.IO import xmlmenudump
from oof2.common import registeredclass
from oof2.engine import mesh
from oof2.engine import skeletoncontext
from oof2.engine import skeletonmodifier
from oof2.engine.IO import outputDefs
import oof2.SWIG.engine.material

FloatRangeParameter = parameter.FloatRangeParameter
IntRangeParameter = parameter.IntRangeParameter
AutoNumericParameter = parameter.AutoNumericParameter

##################

# Add a parameter to the graphics window's Settings menu, to control
# whether or not Mesh Elements with no Material are displayed.

def toggleEmptyDrawing(menuitem, hideEmpty):
    gfxwindow = menuitem.data
    gfxwindow.settings.hideEmptyElements = hideEmpty
    gfxwindow.draw()

def addToGfxSettings(gfxwindow):
    item = gfxwindow.menu.Settings.addItem(oofmenu.CheckOOFMenuItem(
        'Hide_Empty_Mesh_Elements',
        callback=toggleEmptyDrawing,
        value=gfxwindow.settings.hideEmptyElements,
        help="Toggle the display of elements with no Material.",
        discussion=xmlmenudump.loadFile('DISCUSSIONS/common/menu/hideempty.xml')))
    item.data = gfxwindow

switchboard.requestCallback('open graphics window', addToGfxSettings)
ghostgfxwindow.defineGfxSetting('hideEmptyElements', True)

##################

# Skeleton and Mesh display methods are very similar, except for how
# they get some of their data.  These two base classes encapsulate the
# *differences* between Skeletons and Meshes, as far as displaying is
# concerned.  They're mixed in with other base classes to create the
# actual DisplayMethods.

class SkeletonDisplayMethod(display.DisplayMethod):
    def __init__(self):
        display.DisplayMethod.__init__(self)
    def polygons(self, gfxwindow, themesh):
        return [el.perimeter() for el in themesh.element_iterator()]
    
class MeshDisplayMethod(display.DisplayMethod):
    # A display method that displays data from a Mesh at positions
    # determined by the mesh.  self.where is a PositionOutput.
    def __init__(self):
        display.DisplayMethod.__init__(self)
    def incomputable(self, gfxwindow):
        themesh = self.who().resolve(gfxwindow)
        return display.DisplayMethod.incomputable(self, gfxwindow) \
               or self.where.incomputable(themesh)
    def polygons(self, gfxwindow, themesh):
        if gfxwindow.settings.hideEmptyElements:
            edges = [element.perimeter()
                     for element in themesh.element_iterator()
                     if element.material() is not None]
        else:
            edges = [element.perimeter()
                     for element in themesh.element_iterator()]
        flatedges = utils.flatten(edges)
        # corners tells where on each edge to evaluate self.where
        corners = [[0.0]]*len(flatedges)
        # evaluate position output for all edges at once
        pgons = self.where.evaluate(themesh, flatedges, corners)
        # give the corner list the same structure as the edge list: a
        # list of lists, where each sublist is the list of corners of
        # an element.
        pgons = utils.unflatten(edges, pgons)
        if len(pgons) == 0:
#            reporter.warn("No mesh elements drawn! Are there no materials assigned?")
            mainthread.runBlock(
                reporter.warn,
                ("No mesh elements drawn! Are there no materials assigned?",))
        return pgons

###########################
    
class EdgeDisplay:
    def draw(self, gfxwindow, device):
        themesh = self.who().getObject(gfxwindow)
        polygons = self.polygons(gfxwindow, themesh)
        device.comment("EdgeDisplay")
        device.set_lineColor(self.color)
        device.set_lineWidth(self.width)
        for polygonset in polygons:
            device.draw_polygon(primitives.Polygon(polygonset))

class MeshEdgeDisplay(EdgeDisplay, MeshDisplayMethod):
    # EdgeDisplay draws the edges of the Elements
    def __init__(self, where, width=0, color=color.black):
        if where is not None:
            self.where = where.clone()
        else:
            self.where = None           # ?
        self.width = width
        self.color = color
        MeshDisplayMethod.__init__(self)
    def draw(self, gfxwindow, device):
        EdgeDisplay.draw(self, gfxwindow, device)
    def getTimeStamp(self, gfxwindow):
        return max(
            MeshDisplayMethod.getTimeStamp(self, gfxwindow),
            gfxwindow.settings.getTimeStamp('hideEmptyElements'))

class SkeletonEdgeDisplay(EdgeDisplay, SkeletonDisplayMethod):
    def __init__(self, width=0, color=color.black):
        self.width = width
        self.color = color
        SkeletonDisplayMethod.__init__(self)

registeredclass.Registration('Element Edges',
                             display.DisplayMethod,
                             MeshEdgeDisplay,
                             ordering=0.0,
                             layerordering=display.Linear,
                             params=[
    output.PositionOutputParameter('where', outputDefs.originalPosition,
                            tip="Plot at displaced or original position?"),
    color.ColorParameter('color', color.black,
                         tip="Color of the displayed edges."),
    IntRangeParameter('width', (0,10), 0, tip="Line thickness, in pixels.")],
                             whoclasses = ('Mesh',),
                             tip="Draw the edges of Mesh Elements.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/meshedgedisplay.xml')
                             )
    
registeredclass.Registration('Element Edges',
                             display.DisplayMethod,
                             SkeletonEdgeDisplay,
                             ordering=0.0,
                             layerordering=display.Linear,
                             params=[
    color.ColorParameter('color', color.black,
                         tip="Color of the displayed edges."),
    IntRangeParameter('width', (0,10), 0, tip="Line thickness, in pixels.")],
                             whoclasses = ('Skeleton',),
                             tip="Draw the edges of Skeleton Elements.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/skeletonedgedisplay.xml')
                             )

######################

class PerimeterDisplay(MeshDisplayMethod):
    def __init__(self, where, width=0, color=color.black):
        self.where = where.clone()
        self.width = width
        self.color = color
        MeshDisplayMethod.__init__(self)
    def draw(self, gfxwindow, device):
        themesh = self.who().resolve(gfxwindow)
        femesh = themesh.getObject()
        device.comment("PerimeterDisplay")
        device.set_lineColor(self.color)
        device.set_lineWidth(self.width)
        for element in femesh.element_iterator():
            el_edges = element.perimeter()
            for edge in el_edges:
                if element.exterior(edge.startpt(), edge.endpt()):
                    pts = self.where.evaluate(femesh, [edge], [[0.0, 1.0]])
                    device.draw_segment(primitives.Segment(pts[0], pts[1]))

if devel.devel()>=1:
    # This doesn't work as expected...
    registeredclass.Registration('Perimeter',
                                 display.DisplayMethod,
                                 PerimeterDisplay,
                                 ordering=1.0,
                                 layerordering=display.SemiLinear,
                                 params=[
        output.PositionOutputParameter('where', outputDefs.originalPosition,
                                tip="Plot at displaced or original position?"),
        color.ColorParameter('color', color.black,
                             tip=parameter.emptyTipString),
        IntRangeParameter('width', (0,10), 0,
                          tip="Line width.")
        ],
                                 whoclasses = ('Mesh',),
                                 tip="Outline the perimeter of the Mesh",
                                 discussion=xmlmenudump.loadFile(
        'DISCUSSIONS/engine/reg/perimeterdisplay.xml')
                 )

######################

##class CenterFillDisplay:
##    # CenterFillDisplay fills each Element with a solid color,
##    # determined by the 'what' at the center of the Element.

##    def incomputable(self, gfxwindow):
##        themesh = self.who().resolve(gfxwindow)
##        return self.what.incomputable(themesh) or \
##               MeshDisplayMethod.incomputable(self, gfxwindow)
##    def draw(self, gfxwindow, device):
##        themesh = self.who().resolve(gfxwindow).mesh()
##        device.comment("SolidFill")
##        device.set_colormap(self.colormap)
##        polygons = self.polygons(themesh)
##        elements = tuple(themesh.element_iterator())
##        evaluationpoints = [[el.center()] for el in elements]
##        values = self.what.evaluate(themesh, elements, evaluationpoints)
##        max_value = max(values)
##        min_value = min(values)
##        # HACK: If the values are equal, put the actual value in
##        # the middle of an arbitrary range.
##        if max_value==min_value:
##            max_value += 1.0
##            min_value -= 1.0
##        for polygon, value in zip(polygons, values):
##            if value is not None:
##                device.set_fillColor((value-min_value)/(max_value-min_value))
##                device.fill_polygon(primitives.Polygon(polygon))
        
##class MeshCenterFillDisplay(CenterFillDisplay, MeshDisplayMethod): 
##    def __init__(self, where, what, colormap=colormap.ThermalMap()):
##        self.where = where.clone()
##        self.what = what.clone()
##        self.colormap = colormap
##        MeshDisplayMethod.__init__(self)


##registeredclass.Registration('Solid Fill',
##                             display.DisplayMethod,
##                             MeshCenterFillDisplay,
##                             ordering=2.0,
##                             layerordering=display.Planar,
##                             params=[
##    output.ScalarOutputParameter('what', tip='Quantity to be plotted'),
##    output.PositionOutputParameter('where', outputDefs.originalPosition,
##                            tip='Plot at original or displaced position?'),
##    Parameter('colormap', colormap.ColorMap, colormap.ThermalMap(),
##              tip="Fill color.")],
##                             whoclasses = ('Mesh',),
##                             tip="Quick and dirty contour plot.",
##                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/centerfilldisplay.xml')
##                             )

# SkeletonCenterFillDisplay isn't useful unless there are Outputs that
# can be evaluated on a Skeleton.

##class SkeletonCenterFillDisplay(CenterFillDisplay, SkeletonDisplayMethod): 
##    def __init__(self, what, colormap=colormap.ThermalMap()):
##        self.what = what
##        self.colormap = colormap

##registeredclass.Registration('SolidFill',
##             display.DisplayMethod,
##             SkeletonCenterFillDisplay,
##             ordering=2.01,
##             params=[
##    ScalarOutputParameter('what', tip='Quantity to be plotted'),
##    Parameter('colormap', colormap.ColorMap, colormap.ThermalMap())],
##             whoclasses = ('Mesh')
##             )

######################

class MaterialDisplay:
    def draw(self, gfxwindow, device):
        device.comment("Material Color")
        themesh = self.who().resolve(gfxwindow).getObject()
        polygons = self.polygons(gfxwindow, themesh)
        # colorcache is a dictionary of colors keyed by Material.  It
        # prevents us from having to call material.fetchProperty for
        # each element.
        colorcache = {}
        for polygon, material in zip(polygons,
                                     self.materials(gfxwindow, themesh)):
            if material is not None:
                try:
                    # If material has been seen already, retrieve its color.
                    color = colorcache[material]
                except KeyError:
                    # This material hasn't been seen yet.
                    try:
                        colorprop = material.fetchProperty('Color')
                        color = colorprop.color()
                    except ooferror.ErrNoSuchProperty:
                        color = None
                    colorcache[material] = color
                if color is not None:
                    device.set_fillColor(color)
                    device.fill_polygon(primitives.Polygon(polygon))
    def getTimeStamp(self, gfxwindow):
        microstructure = self.who().resolve(gfxwindow).getMicrostructure()
        return max(display.DisplayMethod.getTimeStamp(self, gfxwindow),
                   oof2.SWIG.engine.material.getMaterialTimeStamp(microstructure))
                    
class SkeletonMaterialDisplay(MaterialDisplay, SkeletonDisplayMethod):
    def __init__(self):
        SkeletonDisplayMethod.__init__(self)
    def materials(self, gfxwindow, themesh):
        return [element.material(themesh)
                for element in themesh.element_iterator()]

class MeshMaterialDisplay(MaterialDisplay, MeshDisplayMethod):
    def __init__(self, where):
        self.where = where.clone()
        MeshDisplayMethod.__init__(self)
    def materials(self, gfxwindow, themesh):
        # Because MeshDisplayMethod.polygons only returns the borders
        # of elements with an assigned material, this should only
        # return the non-trivial materials.
        allmats = [element.material()
                   for element in themesh.element_iterator()]
        if gfxwindow.settings.hideEmptyElements:
            return filter(None, allmats)
        return allmats
    def getTimeStamp(self, gfxwindow):
        return max(MaterialDisplay.getTimeStamp(self, gfxwindow),
                   gfxwindow.settings.getTimeStamp('hideEmptyElements'))
    
    
registeredclass.Registration('Material Color',
                             display.DisplayMethod,
                             SkeletonMaterialDisplay,
                             layerordering=display.Planar(1),
                             ordering=0.1,
                             params=[],
                             whoclasses=('Skeleton',),
                             tip="Fill each Element with the color of its assigned Material.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/skeletonmaterialdisplay.xml')
                             )

registeredclass.Registration('Material Color',
                             display.DisplayMethod,
                             MeshMaterialDisplay,
                             ordering=0.11,
                             layerordering=display.Planar(2),
                             params=[
    output.PositionOutputParameter('where', outputDefs.originalPosition,
                           tip='Plot at original or displaced position?')],
                             whoclasses=('Mesh',),
                             tip="Fill each Element with the color of its assigned Material.",
                                                          discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/meshmaterialdisplay.xml')
                             )

###########################################

class SkeletonQualityDisplay(SkeletonDisplayMethod):
    contourmaplevels = 32
    def __init__(self, alpha, min, max, colormap):
        self.alpha = alpha
        self.colormap = colormap
        self.max = max
        self.min = min
        self.vmax = None
        self.vmin = None
        self.contourmaphidden = 0
        self.lock = lock.Lock()
        SkeletonDisplayMethod.__init__(self)
    def draw(self, gfxwindow, device):
        self.lock.acquire()
        try:
            device.comment("Skeleton Energy")
            skel = self.who().resolve(gfxwindow).getObject()
            # get polygons and element energy in one pass
            polyenergy = [(el.perimeter(), el.energyTotal(skel, self.alpha))
                        for el in skel.element_iterator()
                        if not el.illegal()]
            # find actual range of data
            self.vmax = self.vmin = polyenergy[0][1]
            for (p,e) in polyenergy[1:]:
                if e > self.vmax:
                    self.vmax = e
                if e < self.vmin:
                    self.vmin = e
            # Set plot limits to either the actual data extremes, or
            # to the passed in values.  Store the actual limits in
            # vmin and vmax.
            if self.max == automatic.automatic:
                max = self.vmax
            else:
                max = self.max
                self.vmax = max
            if self.min == automatic.automatic:
                min = self.vmin
            else:
                min = self.min
                self.vmin = min
            if max == min:
                max += 1.0
                self.vmax += 1.0
                min -= 1.0
                self.vmin -= 1.0
            device.set_colormap(self.colormap)
            for polygon, energy in polyenergy:
                device.set_fillColor((energy-min)/(max-min))
                device.fill_polygon(primitives.Polygon(polygon))
        finally:
            self.lock.release()
    def getTimeStamp(self, gfxwindow):
        skelcontext = self.who().resolve(gfxwindow)
        return max(self.timestamp,
                   skelcontext.getTimeStamp(gfxwindow),
                   skelcontext.getMicrostructure().getTimeStamp())
    def contour_capable(self, gfxwindow):
        return not self.incomputable(gfxwindow)
    def get_contourmap_info(self):
        if self.vmax is not None:
            delta = (self.vmax - self.vmin)/(self.contourmaplevels-1.)
            return (self.vmin, self.vmax,
                    [self.vmin+x*delta for x in range(self.contourmaplevels)])
        return (0., 1., [0])
    def hide_contourmap(self):
        self.contourmaphidden = 1
    def show_contourmap(self):
        self.contourmaphidden = 0
    def draw_contourmap(self, gfxwindow, device):
        self.lock.acquire()
        try:
            if self.vmax is not None:
                aspect_ratio = gfxwindow.settings.aspectratio
                height = self.vmax - self.vmin
                width = height/aspect_ratio
                delta = height/(self.contourmaplevels-1.)
                device.comment("Colorbar minimum: %g" % self.vmin)
                device.comment("Colorbar maximum: %g" % self.vmax)
                device.set_colormap(self.colormap)
                for i in range(self.contourmaplevels):
                    low = i*delta
                    high = (i+1)*delta
                    rect_bdy = [primitives.Point(0.0, low),
                                primitives.Point(0.0, high),
                                primitives.Point(width, high),
                                primitives.Point(width, low)]
                    rectangle = primitives.Polygon(rect_bdy)
                    if height > 0.0:
                        device.set_fillColor(low/height)
                    else:
                        device.set_fillColor(0.0)
                    device.fill_polygon(rectangle)
        finally:
            self.lock.release()
    

registeredclass.Registration(
    'SkeletonQuality',
    display.DisplayMethod,
    SkeletonQualityDisplay,
    ordering=100,
    layerordering=display.Planar,
    whoclasses=('Skeleton',),
    params=[skeletonmodifier.alphaParameter,
            parameter.RegisteredParameter('colormap', colormap.ColorMap,
                                          colormap.ThermalMap(),
                                          tip="color scheme"),
            AutoNumericParameter('min', automatic.automatic,
                               tip="lowest energy to display, or 'automatic'"),
            AutoNumericParameter('max', automatic.automatic,
                              tip="highest energy to display, or 'automatic'")
            ],
    tip="Color each element according to its effective energy.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/skelqualdisplay.xml')
    )

###########################################

ghostgfxwindow.DefaultLayer(skeletoncontext.skeletonContexts,
                            SkeletonEdgeDisplay())

ghostgfxwindow.DefaultLayer(mesh.meshes,
                            MeshEdgeDisplay(where=outputDefs.actualPosition))
