# -*- python -*-
# $RCSfile: skeletonIO.py,v $
# $Revision: 1.59 $
# $Author: langer $
# $Date: 2006/02/02 20:51:06 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Menu commands for reading a Skeleton from a file.

# TODO LATER: the name of the Skeleton appears many times in the file,
# in each menu command.  That makes it hard for the user to edit the
# file to change the name.  Ugly solution: have a global
# "currentskeleton" variable which is set by the initial Skeleton.New
# command.  Difficult solution: allow variables to be set locally *in*
# the data file, and store the skeleton name there.

from oof2.SWIG.common import switchboard
from oof2.common import debug
from oof2.common.IO import mainmenu
from oof2.common.IO import oofmenu
from oof2.common.IO import parameter
from oof2.common.IO import whoville
from oof2.engine import skeletonboundary
from oof2.engine import skeletoncontext
import oof2.common.microstructure
import oof2.engine.skeleton
from oof2.SWIG.common import devel

OOFMenuItem = oofmenu.OOFMenuItem

OOF = mainmenu.OOF

skelmenu = OOF.LoadData.addItem(OOFMenuItem('Skeleton'))

def _newSkel(menuitem, name, microstructure):
    oof2.engine.skeleton.newEmptySkeleton(name, microstructure)
    
skelmenu.addItem(OOFMenuItem(
    'New',
    callback=_newSkel,
    params=[parameter.StringParameter('name', tip="Name for the Skeleton."),
            whoville.WhoParameter('microstructure',
                         oof2.common.microstructure.microStructures,
                                  tip=parameter.emptyTipString)],
    help="Load Skeleton. Used internally in a data file.",
    discussion="<para>Initiate loading &skel; from a data file.</para>"))

###

## The order of the nodes in a Skeleton must not change when the
## Skeleton is saved and reloaded, or the Field values on an
## associated Mesh will be assigned to the wrong nodes. See comment in
## meshIO.py.

def _loadNodes(menuitem, skeleton, points):
    # read nodes as (x,y) tuples of floats
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    for node in points:
        skeleton.newNode(*node)
    switchboard.notify(('who changed', 'Skeleton'), skelcontext)

skelmenu.addItem(OOFMenuItem(
    'Nodes',
    callback=_loadNodes,
    params=[whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                                  tip=parameter.emptyTipString),
            parameter.ListOfTuplesOfFloatsParameter('points',
                                                    tip="List of points (nodes).")],
    help="Load Nodes. Used internally in a data file.",
    discussion="<para>Load <link linkend='Section:Concepts:Skeleton:Node'>nodes</link> from a <link linkend='MenuItem:OOF.File.Save.Skeleton'>saved</link> Skeleton.</para>"
    ))

###

def _loadElements(menuitem, skeleton, nodes):
    # read elements as tuples of node indices
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    for nodelist in nodes:
        skeleton.loadElement(*nodelist)
    skelcontext.getTimeStamp(None).increment()
    switchboard.notify(('who changed', 'Skeleton'), skelcontext)

skelmenu.addItem(OOFMenuItem(
    'Elements',
    callback=_loadElements,
    params=[whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                                  tip=parameter.emptyTipString),
            parameter.ListOfTuplesOfIntsParameter('nodes',
                                                  tip="List of element connectivities (List of node indices).")],
    help="Load Elements. Used internally in a data file.",
    discussion="<para>Load <link linkend='Section:Concepts:Skeleton:Element'>elements</link> from a <link linkend='MenuItem:OOF.File.Save.Skeleton'>saved</link> Skeleton.</para>"
    ))

###

def _loadPinnedNodes(menuitem, skeleton, nodes):
    # "nodes" is a list of integer indices into skeleton.nodes
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    pinned = skelcontext.pinnednodes
    pinned.pin([skeleton.nodes[i] for i in nodes])
    pinned.signal()
skelmenu.addItem(OOFMenuItem(
    'PinnedNodes',
    callback = _loadPinnedNodes,
    params = [
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                          tip=parameter.emptyTipString),
    parameter.ListOfIntsParameter('nodes', tip="List of indices of pinned nodes.")],
    help="Load pinned Nodes. Used internally in a data file.",
    discussion="<para>Load <link linkend='MenuItem:OOF.Graphics_n.Toolbox.Pin_Nodes.Pin'>pinned</link> nodes from a &skel; data file.</para>"
    ))

###
    
def _loadPointBoundary(menuitem, skeleton, name, nodes, exterior):
    # "nodes" is a list of integer indices into skeleton.nodes
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    skelcontext.createPointBoundary(name, [skeleton.nodes[i] for i in nodes],
                                    exterior, autoselect=0)
    # autoselect=0 prevents the boundary from being selected in the
    # skelcontext. Selection shouldn't happen except by explicit user
    # action.

skelmenu.addItem(OOFMenuItem(
    'PointBoundary',
    callback=_loadPointBoundary,
    params=[
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                          tip=parameter.emptyTipString),
    parameter.StringParameter('name', tip="Name of Point Boundary."),
    parameter.ListOfIntsParameter('nodes', tip="List of node indices."),
    parameter.IntParameter('exterior', 0, tip="1 (true) for the exterior boundary and 0 (false) for otherwise.")],
    help="Load Point Boundary. Used internally in a data file.",
    discussion="<para>Load a <link linkend='Section:Concepts:Skeleton:Boundary:Point'><classname>PointBoundary</classname></link> from a Skeleton data file.</para>"
    ))

###

def _loadEdgeBoundary(menuitem, skeleton, name, edges, exterior):
    # "edges" is a list of pairs of integer indices into skeleton.nodes
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    seglist = [skeleton.getSegment(
        *[skeleton.nodes[i] for i in e] )
               for e in edges]
    startnode = skeleton.nodes[edges[0][0]]
    skelcontext.createEdgeBoundary(name, seglist, startnode,
                                   exterior, autoselect=0)
    # autoselect=0 prevents the boundary from being selected in the
    # skelcontext. Selection shouldn't happen except by explicit user
    # action.

    
skelmenu.addItem(OOFMenuItem(
    'EdgeBoundary',
    callback=_loadEdgeBoundary,
    params=[
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                          tip=parameter.emptyTipString),
    parameter.StringParameter('name', tip="Name of Edge Boundary."),
    parameter.ListOfTuplesOfIntsParameter('edges',
                                          tip="List of Edges -- tuple of two nodes."),
    parameter.IntParameter('exterior', 0, tip="1 (true) for the exterior boundary and 0 (false) for otherwise.")],
    help="Load Edge Boundary. Used internally in a data file.",
    discussion="<para>Load a <link linkend='Section:Concepts:Skeleton:Boundary:Edge'><classname>EdgeBoundary</classname></link> from a Skeleton data file.</para>"
    ))

######

def _loadNodeGroup(menuitem, skeleton, name, nodes):
    # nodes is a list of integer indices into skeleton.nodes.
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    groupset = skelcontext.nodegroups
    groupset.addGroup(name)
    groupset.addToGroup(name, [skeleton.nodes[i] for i in nodes])

skelmenu.addItem(OOFMenuItem(
    'NodeGroup',
    callback=_loadNodeGroup,
    params=[
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                          tip=parameter.emptyTipString),
    parameter.StringParameter('name', tip="Name for the node group."),
    parameter.ListOfIntsParameter('nodes', tip="List of node indices.")],
    help="Load Node Group. Used internally in a data file.",
    discussion="<para>Load a node <link linkend='Section:Concepts:Skeleton:Groups'>group</link>.</para>"
    ))

def _loadElementGroup(menuitem, skeleton, name, elements):
    # elements is a list of integer indices into skeleton.elements.
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    groupset = skelcontext.elementgroups
    groupset.addGroup(name)
    groupset.addToGroup(name, [skeleton.elements[i] for i in elements])

skelmenu.addItem(OOFMenuItem(
    'ElementGroup',
    callback=_loadElementGroup,
    params=[
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                          tip=parameter.emptyTipString),
    parameter.StringParameter('name', tip="Name for the element group."),
    parameter.ListOfIntsParameter('elements', tip="List of element indices")],
    help="Load Element Group. Used internally in a data file.",
    discussion="<para>Load an element <link linkend='Section:Concepts:Skeleton:Groups'>group</link>.</para>"
    ))

def _loadSegmentGroup(menuitem, skeleton, name, segments):
    # segments is a list of tuples of integer indices into skeleton.nodes.
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    groupset = skelcontext.segmentgroups
    groupset.addGroup(name)
    groupset.addToGroup(name, [skeleton.getSegment(skeleton.nodes[i],
                                                   skeleton.nodes[j])
                               for (i,j) in segments])

skelmenu.addItem(OOFMenuItem(
    'SegmentGroup',
    callback=_loadSegmentGroup,
    params=[
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                          tip=parameter.emptyTipString),
    parameter.StringParameter('name', tip="Name for the segment group."),
    parameter.ListOfTuplesOfIntsParameter('segments', tip="List of segments -- tuple of two node indices.")],
    help="Load Segment Group. Used internally in a data file.",
    discussion="<para>Load a segment <link linkend='Section:Concepts:Skeleton:Groups'>group</link>.</para>"
    ))


#############

def writeSkeleton(datafile, skelcontext):
    skelcontext.begin_reading()
    try:
        skeleton = skelcontext.getObject()
        skelpath = skelcontext.path()

        # Create skeleton.
        datafile.startCmd(skelmenu.New)
        datafile.argument('name', skelcontext.name())
        datafile.argument('microstructure', skeleton.MS.name())
        datafile.endCmd()

        # Define nodes.
        datafile.startCmd(skelmenu.Nodes)
        datafile.argument('skeleton', skelpath)
        datafile.argument('points', [(nd.position().x, nd.position().y)
                                     for nd in skeleton.nodes])
        datafile.endCmd()

        # Nodes are written and read in the order in which they're stored
        # in the Skeleton's nodes list.  This is not the same as the
        # nodes' internal index, so we need to create a dictionary of
        # indices in the nodes list so that the elements' node tuples are
        # correct.
        nodedict = {}
        c = 0
        for node in skeleton.nodes:
            nodedict[node] = c
            c += 1
        # same for elements
        elementdict = {}
        c = 0
        for element in skeleton.elements:
            elementdict[element] = c
            c += 1

        # Define elements, using the node indices in nodedict.
        datafile.startCmd(skelmenu.Elements)
        datafile.argument('skeleton', skelpath)
        datafile.argument('nodes', [tuple([nodedict[node] for node in el.nodes])
                                    for el in skeleton.elements])
        datafile.endCmd()

        # Node groups
        for group in skelcontext.nodegroups.groups:
            datafile.startCmd(skelmenu.NodeGroup)
            datafile.argument('skeleton', skelpath)
            datafile.argument('name', group)
            datafile.argument('nodes', [nodedict[node] for node in skelcontext.nodegroups.get_group(group)])
            datafile.endCmd()

        # Element groups
        for group in skelcontext.elementgroups.groups:
            datafile.startCmd(skelmenu.ElementGroup)
            datafile.argument('skeleton', skelpath)
            datafile.argument('name', group)
            datafile.argument(
                'elements',
                [elementdict[el]
                 for el in skelcontext.elementgroups.get_group(group)])
            datafile.endCmd()

        for group in skelcontext.segmentgroups.groups:
            datafile.startCmd(skelmenu.SegmentGroup)
            datafile.argument('skeleton', skelpath)
            datafile.argument('name', group)
            nodepairs = [seg.nodes()
                         for seg in skelcontext.segmentgroups.get_group(group)]
            datafile.argument('segments',
                              [(nodedict[n1], nodedict[n2]) for (n1, n2) in nodepairs])
            datafile.endCmd()

        # Pinned nodes
        datafile.startCmd(skelmenu.PinnedNodes)
        datafile.argument('skeleton', skelpath)
        datafile.argument('nodes', [nodedict[node] for node in skelcontext.pinnednodes.retrieve()])
        datafile.endCmd()

        # Point boundaries
        for pbname, pbdy in skeleton.pointboundaries.items():
            datafile.startCmd(skelmenu.PointBoundary)
            datafile.argument('skeleton', skelpath)
            datafile.argument('name', pbname)
            datafile.argument('nodes', [nodedict[node] for node in pbdy.nodes])
            exterior = 0
            if isinstance(pbdy, skeletonboundary.ExteriorSkeletonPointBoundary):
                exterior = 1
            datafile.argument('exterior', exterior)
            datafile.endCmd()

        # Edge boundaries
        for ebname, ebdy in skeleton.edgeboundaries.items():
            datafile.startCmd(skelmenu.EdgeBoundary)
            datafile.argument('skeleton', skelpath)
            datafile.argument('name', ebname)
            edgeset = rearrangeEdges([
                tuple([nodedict[x] for x in edge.get_nodes()])
                for edge in ebdy.edges
                ])
            datafile.argument('edges', edgeset)
            exterior = 0
            if isinstance(ebdy, skeletonboundary.ExteriorSkeletonEdgeBoundary):
                exterior = 1
            datafile.argument('exterior', exterior)
            datafile.endCmd()
    finally:
        skelcontext.end_reading()

##    # debug purpose code.
##    file = open("temp", "w")
##    file.write("Skeleton.Nodes(skeleton='hole.png:skeleton', points=[")
##    nodes = []
##    for nd in skeleton.node_iterator():
##        neighbors = nd.neighborElements()
##        for el in neighbors:
##            if el.dominantPixel(skeleton)==0:
##                file.write("(%g,%g)" % (nd.position().x, nd.position().y))
##                file.write(", ")
##                nodes.append(nd)
##                break
##    file.write("])\n")

    

##    file.write("Skeleton.Elements(skeleton='el_shape.png:skeleton', nodes=[")
##    for el in skeleton.element_iterator():
##        if el.dominantPixel(skeleton)==0:
##            if el.nnodes() == 4:
##                file.write("(%d, %d, %d, %d)" % (nodes.index(el.nodes[0]),
##                                                 nodes.index(el.nodes[1]),
##                                                 nodes.index(el.nodes[2]),
##                                                 nodes.index(el.nodes[3])))
##            else:
##                file.write("(%d, %d, %d)" % (nodes.index(el.nodes[0]),
##                                             nodes.index(el.nodes[1]),
##                                             nodes.index(el.nodes[2])))
##            file.write(", ")
##    file.write("])\n")
##    file.close()
        
###########################################################

# A function to put edges in a right order.
def rearrangeEdges(edges):
    # 'edges' is a list of tuples of integers.  Each tuple is an edge,
    # and each integer is a node index (from nodedict in
    # writeSkeleton, above).
    if len(edges) == 1:
        return edges

    # Construct dictionaries of edges keyed by their starting and end points.
    tails = {}           
    heads = {}         
    for edge in edges:
        heads[edge[0]] = edge
        tails[edge[1]] = edge

    newedges = [edges[0]]

    # Build the list of contiguous edges going backwards from the
    # start point of edges[0].
    lastpt = edges[0][0]
    try:
        while 1:
            next = tails[lastpt]
            newedges.append(next)
            lastpt = next[0]
    except KeyError:
        pass

    newedges.reverse()

    # Extend the list of contiguous edges going forwards from the end
    # point of edges[0].
    lastpt = edges[0][1]
    try:
        while 1:
            next = heads[lastpt]
            newedges.append(next)
            lastpt = next[1]
    except KeyError:
        pass

    return newedges


##########
# ABAQUS #
##########
import datetime
import string
def writeABAQUSfromSkeleton(filename, mode, skelcontext):
    skelcontext.begin_reading()
    try:
        skeleton = skelcontext.getObject()

        buffer="*HEADING\nABAQUS-style file created by OOF2 on %s from a skeleton " % (datetime.datetime.today())
        buffer+="of the microstructure %s.\n" % skeleton.MS.name()

        # Build dictionary (instead of using index()) for elements and nodes
        #  as was done in previous writeXXX() methods
        nodedict = {}
        i = 1
        for node in skeleton.nodes:
            nodedict[node] = i
            i += 1
        # same for elements
        elementdict = {}
        i = 1
        for el in skeleton.elements:
            elementdict[el] = i
            i += 1

        # Collect elements with the same dominant material together in a
        #  dictionary, with a key given by the material name.
        #  Some elements may not have a material assigned and
        #  these should not be included in the dictionary(?). Have a feeling
        #  something like this has been done in the OOF universe.
        materiallist={}
        elementlist={}
        for el in skeleton.elements:
            if el.material(skeleton):
                matname=el.material(skeleton).name()
                # Steve's suggestion
                elindex = elementdict[el]
                try:
                    elementlist[matname].append(elindex)
                except KeyError:
                    elementlist[matname] = [elindex]
                    materiallist[matname] = el.material(skeleton)

        buffer+="** Materials defined by OOF2:\n"
        for matname, details in materiallist.items():
            buffer+="**   %s:\n" % (matname)
            for prop in details.properties():
                for param in prop.registration().params:
                    buffer+="**     %s: %s\n" % (param.name,param.value)

        buffer+="** Notes:\n**   The nodes for a skeleton are always located at vertices or corners.\n"
        buffer+="**   More information may be obtained by saving ABAQUS from a mesh.\n"

        listbuf=["*NODE\n"]
        for node in skeleton.nodes:
            listbuf.append("%d, %s, %s\n" % (nodedict[node],node.position().x,node.position().y))
        buffer+=string.join(listbuf,"")

        # Only expecting 3 or 4 noded skeleton elements
        for numnodes in [3,4]:
            listbuf=["** The element type provided for ABAQUS is only a guess " \
                     "and may have to be modified by the user to be meaningful.\n*ELEMENT, TYPE=CPS%d\n" % numnodes]
            for el in skeleton.elements:
                if el.nnodes()==numnodes:
                    listbuf2=["%d" % (elementdict[el])]
                    for node in el.nodes:
                        listbuf2.append("%d" % (nodedict[node]))
                    listbuf.append(string.join(listbuf2,", ")+"\n")
            if len(listbuf)>1:
                buffer+=string.join(listbuf,"")

        for group in skelcontext.nodegroups.groups:
            buffer+="*NSET, NSET=%s\n" % (group)
            listbuf=[]
            i=0
            for node in skelcontext.nodegroups.get_group(group):
                if i>0 and i%16==0:
                    listbuf.append("\n%d" % (nodedict[node]))
                else:
                    listbuf.append("%d" % (nodedict[node]))
                i+=1
            buffer+=string.join(listbuf,", ")+"\n"

        for elgroup in skelcontext.elementgroups.groups:
            buffer+="*ELSET, ELSET=%s\n" % (elgroup)
            listbuf=[]
            i=0
            for el in skelcontext.elementgroups.get_group(elgroup):
                if i>0 and i%16==0:
                    listbuf.append("\n%d" % (elementdict[el]))
                else:
                    listbuf.append("%d" % (elementdict[el]))
                i+=1
            buffer+=string.join(listbuf,", ")+"\n"

        buffer+="** Include point and edge boundaries from OOF2.\n"
        for pbname, pbdy in skeleton.pointboundaries.items():
            buffer+="*NSET, NSET=%s\n" % (pbname)
            listbuf=[]
            i=0
            for node in pbdy.nodes:
                if i>0 and i%16==0:
                    listbuf.append("\n%d" % (nodedict[node]))
                else:
                    listbuf.append("%d" % (nodedict[node]))
                i+=1
            buffer+=string.join(listbuf,", ")+"\n"

        # Use rearrangeEdges() to chain the edges together, then pick the
        #  unique nodes. It seems the edges can't be selected if they
        #  are empty, so edgeset=[(a,b),(b,c),...] is not checked
        #  for null content
        for ebname, ebdy in skeleton.edgeboundaries.items():
            edgeset = rearrangeEdges([
                tuple([nodedict[node] for node in edge.get_nodes()])
                for edge in ebdy.edges
                ])
            buffer+="*NSET, NSET=%s\n" % (ebname)
            listbuf=["%d" % edgeset[0][0]]
            i=1
            for edge in edgeset:
                if i%16==0:
                    listbuf.append("\n%d" % (edge[1]))
                else:
                    listbuf.append("%d" % (edge[1]))
                i+=1
            buffer+=string.join(listbuf,", ")+"\n"

        for matname in materiallist.keys():
            buffer+="*ELSET, ELSET=%s\n" % matname
            listbuf=[]
            i=0
            for elindex in elementlist[matname]:
                if i>0 and i%16==0:
                    listbuf.append("\n%d" % (elindex))
                else:
                    listbuf.append("%d" % (elindex))
                i+=1
            buffer+=string.join(listbuf,", ")+"\n*SOLID SECTION, ELSET=%s, MATERIAL=%s\n" % (matname,matname)

        for matname in materiallist.keys():
            buffer+="*MATERIAL, NAME=%s\n** Use the information in the header to complete these fields under MATERIAL\n" % matname

        # Save/Commit to file. Perhaps should be done outside the current method.
        fp=open(filename,mode)
        fp.write(buffer)
        fp.close()
    finally:
        skelcontext.end_reading()
