# -*- python -*-
# $RCSfile: errorestimator.py,v $
# $Revision: 1.16 $
# $Author: langer $
# $Date: 2006/03/02 18:59:54 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


from oof2.SWIG.common.IO import progressbar
from oof2.common import registeredclass
from oof2.common.IO import mainmenu
from oof2.common.IO import parameter
from oof2.common.IO import whoville
from oof2.common.IO import xmlmenudump
from oof2.engine import refinementtarget
from oof2.engine.IO import meshparameters
import oof2.engine.mesh

import random


# ErrorEstimator base class
class ErrorEstimator(registeredclass.RegisteredClass):
    registry = []
    # If an ErrorEstimator needs a preprocess, it should be added.
    def preprocess(self, mesh):
        pass
    tip = "Error estimators."
    discussion = xmlmenudump.loadFile(
        'DISCUSSIONS/engine/reg/errorestimator.xml')

####################################

class ErrorNorm(registeredclass.RegisteredClass):
    registry = []
    def preprocess(self, femesh, flux):  # to be overridden by child class.
        pass
    tip = "How to measure error?"
    discussion = xmlmenudump.loadFile('DISCUSSIONS/engine/reg/errornorm.xml')

## TODO: The ErrorNorms defined here only work for the ZZ error
## estimator.  The classes need to rearranged so that the norm classes
## are independent of the estimator classes.  After all, we may
## someday have more than one estimator.  If it's not possible to
## refactor the classes like that, then the ErrorNorm classes must at
## least be renamed so that it's clear that they are only meant to be
## used with the ZZ estimator.
    
class L2ErrorNorm(ErrorNorm):
    def __call__(self, element, femesh, flux):
        return femesh.zz_L2_estimate(element, flux)
    
registeredclass.Registration(
    'L2 Error Norm',
    ErrorNorm,
    L2ErrorNorm,
    ordering=0,
    tip="Use the root mean square of the components of the error.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/l2errornorm.xml'))

class WeightedL2ErrorNorm(ErrorNorm):
    def __init__(self):
        self.weights = None
    def preprocess(self, femesh, flux):
        bottom = 0.0  # [bottom, top] : the range normalized values
        top = 0.3     # to be weighted.
        self.weights = femesh.zz_L2_weights(flux, bottom, top)
    def __call__(self, element, femesh, flux):
        index = element.get_index()
        return femesh.zz_L2_estimate(element, flux)*self.weights[index]
    
registeredclass.Registration(
    'Weighted L2 Error Norm',
    ErrorNorm,
    WeightedL2ErrorNorm,
    ordering=1,
    tip="Use the weighted root mean square of the components of the error.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/l2weightednorm.xml')
    )

####################################

class ZZ_Estimator(ErrorEstimator):
    def __init__(self, norm, flux, threshold):
        self.norm = norm  # how to normalize at each point
        self.flux = flux  # which flux?
        self.threshold = threshold  # acceptable relative error (percent)
    def preprocess(self, mesh):
        # Superconvergent patch recovery -- Recover flux at each node.
        meshcontext = oof2.engine.mesh.meshes[mesh]
        skel = meshcontext.getSkeleton()
        femesh = meshcontext.femesh()
        femesh.create_scpatch(skel)
        femesh.flux_recovery()
        self.norm.preprocess(femesh, self.flux)
    def __call__(self, element, femesh):
        # Perform an error estimation on the element.
        # FEMesh will take care of the computation.
        return self.threshold < self.norm(element, femesh, self.flux)*100.

# Dummy parameter for smart widget
class ZZFluxParameter(meshparameters.FluxParameter):
    pass

registeredclass.Registration(
    'Z-Z Estimator',
    ErrorEstimator,
    ZZ_Estimator,
    ordering=0.,
    params=[
    parameter.RegisteredParameter('norm',
                                  ErrorNorm,
                                  tip="How to measure the size of an error."),
    ZZFluxParameter('flux', tip=parameter.emptyTipString),
    parameter.FloatRangeParameter('threshold', (0.0, 100.0, 0.5),
                                  value=10.,
                                  tip='Maximum allowable percentage error.')
    ],
    tip="Error estimation by Zienkiewicz and Zhu's superconvergent patch recovery.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/zzspr.xml'))

################################

# RefinementTarget (formerly known as "CheckAndMark")
class AdaptiveMeshRefine(refinementtarget.RefinementTarget):
    def __init__(self, mesh, estimator):
        self.mesh = mesh  # mesh
        self.estimator = estimator  # ErrorEstimator instance
            
    def __call__(self, skeleton, context, divisions, markedEdges, criterion):
        self.estimator.preprocess(self.mesh)
        pBar = progressbar.getProgress()
        femesh = oof2.engine.mesh.meshes[self.mesh].getObject()
        elements = skeleton.activeElements()
        n = len(elements)
        for i in range(n):
            element = elements[i]
            fe_element = femesh.getElement(element.meshindex)
            if self.estimator(fe_element, femesh) and \
               criterion(skeleton, element):
                self.markElement(element, divisions, markedEdges)
            if pBar.query_stop() :
                pBar.set_failure()
                pBar.set_message("Failed")
                break
            else:
                pBar.set_progress(1.0*(i+1)/n)
                pBar.set_message("Performing error estmation: %d/%d" % (i+1, n))

#####################

# A trivial WhoParameter for a smart widget.
# When adaptively refining mesh, "OK" button (in Skeleton Page) should be
# available ONLY when the current skeleton is the base skeleton of
# the to-be-refined mesh.
class AMRWhoParameter(whoville.WhoParameter):
    pass

#####################

registeredclass.Registration(
    'Adaptive Mesh Refinement',
    refinementtarget.RefinementTarget,
    AdaptiveMeshRefine,
    ordering=6,
    params=[AMRWhoParameter("mesh", oof2.engine.mesh.meshes,
                            tip="Apply adaptive refinement to this mesh."),
            parameter.RegisteredParameter("estimator",
                                          ErrorEstimator,
                                          tip="Which error estimator to use.")
    ],
    tip="Refine less accurate elements.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/adaptivemeshrefine.xml'))
