// -*- C++ -*-
// $RCSfile: femesh.C,v $
// $Revision: 1.137 $
// $Author: langer $
// $Date: 2006/03/13 20:26:06 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include <iostream>
#include <string>
#include <math.h>

#include "engine/cscpatch.h"
#include "engine/csrmatrix.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/equation.h"
#include "engine/flux.h"
#include "engine/field.h"
#include "engine/material.h"
#include "engine/femesh.h"
#include "engine/gausspoint.h"
#include "engine/nodalequation.h"
#include "engine/nodalfluxes.h"
#include "engine/nodalscpatches.h"
#include "engine/node.h"
#include "engine/ooferror.h"
#include "engine/outputval.h"
#include "engine/preconditioner.h"
#include "engine/property.h"
#include "engine/smallmatrix.h"
#include "engine/solver.h"
#include "engine/stiffnessmatrix.h"
#include "common/tostring.h"
#include "common/trace.h"
#include "common/printvec.h"
#include "common/switchboard.h"
#include "common/IO/progressbar.h"

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

FEMesh::FEMesh()
  : microstructure(0),\
    rwlock(0),
    precomputeRequired(0),
    ncount(0),
    dof_list_needs_cleaning(false),
    nodaleqn_list_needs_cleaning(false),
    global_equation_map(0),
    n_defined_dof(0),
    n_active_eqn(0),
    n_active_field(0),
    stiffness(0),
    boundary_right_side(0),
    body_force_rhs(0),
    local_time(0),
    c_matrix(0),
    m_matrix(0),
    solve_all(true)
{
}

FEMesh::FEMesh(CMicrostructure * mc)
  : microstructure(mc),
    rwlock(0),
    precomputeRequired(0),
    ncount(0),
    dof_list_needs_cleaning(false),
    nodaleqn_list_needs_cleaning(false),
    global_equation_map(0),
    n_defined_dof(0),
    n_active_eqn(0),
    n_active_field(0),
    stiffness(0),
    boundary_right_side(0),
    body_force_rhs(0),
    local_time(0),
    c_matrix(0),
    m_matrix(0),
    solve_all(true)
{
}

FEMesh::~FEMesh() {
  for(std::vector<NodalEquation*>::size_type i=0; i<nodaleqn.size(); i++)
    delete nodaleqn[i];
  for(std::vector<DegreeOfFreedom*>::size_type i=0; i<dof.size(); i++)
    delete dof[i];
  if(stiffness)
    delete stiffness;
  if(boundary_right_side)
    delete boundary_right_side;
  if(c_matrix)
    delete c_matrix;
  if(m_matrix)
    delete m_matrix;
  // We do not own rwlock, so don't delete it.

  std::map<const int, NodalSCPatches*>::iterator piter;
  for(piter=scpatches.begin(); piter!=scpatches.end(); piter++)
    delete piter->second;

  std::map<const int, NodalFluxes*>::iterator fiter;
  for(fiter=recovered_fluxes.begin(); fiter!=recovered_fluxes.end(); fiter++)
    delete fiter->second;
}

const std::string &FEMesh::classname() const {
  static const std::string _name = "FEMesh";
  return _name;
}

const std::string &FEMesh::modulename() const {
  static const std::string _name = "oof2.SWIG.engine.femesh";
  return _name;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// All Degrees of Freedom are created and deleted with these functions

DegreeOfFreedom *FEMesh::CreateDoF(double x) {
  // localindex is the DoF's position in the node's list of DoFs.
  int index = dof.size();
  DegreeOfFreedom *newdof = new DegreeOfFreedom(x, index);
  dof.push_back(newdof);
  return newdof;
}

void FEMesh::RemoveDoF(DegreeOfFreedom *deaddof) {
//   Trace("FEMesh::RemoveDoF");
  int index = deaddof->dofindex();
  delete deaddof;
  // Don't shift the rest of the vector entries to fill the hole
  // caused by deleting this dof.  The hole can be filled more
  // efficiently by garbage collection later.  Just make sure to check
  // for null pointers when using the dof list!
  dof[index] = 0;
  dof_list_needs_cleaning = true;
}

void FEMesh::clean_doflist() {
  if(dof_list_needs_cleaning) {
    std::vector<DegreeOfFreedom*>::size_type j=0;
    for(std::vector<DegreeOfFreedom*>::size_type i=0; i<dof.size(); i++) {
      if (dof[i] != 0) {
	if(i != j) {
	  dof[j] = dof[i];
	  dof[j]->index_ = j;
	}
	j++;
      }
    }
    if(j != dof.size())
      dof.resize(j);
    dof_list_needs_cleaning=false;
  }
}

// When a Node is created *after* the Fields have been defined, the
// degrees of freedom in the Node need to be activated.  Nodes created
// before the Fields are defined have their degrees of freedom
// activated when the Fields are activated.

void FEMesh::activate_active_dofs(FuncNode *n) const {
  for(std::vector<Field*>::size_type i=0; i<Field::allfields().size(); i++) {
    Field *field = Field::allfields()[i];
    if(field->is_active(this)) {
      for(int k=0; k<field->ndof(); k++) {
	(*field)(this, n, k)->activate();
      }
    }
  }
}

int FEMesh::n_defined_dofs() const {
  return n_defined_dof;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// All NodalEquations are created and deleted with these functions

NodalEquation *FEMesh::CreateNodalEqn() {
//   Trace("FEMesh::CreateNodalEqn");
  int index = nodaleqn.size();
  NodalEquation *neweqn = new NodalEquation(index);
  nodaleqn.push_back(neweqn);
  return neweqn;
}

void FEMesh::RemoveNodalEqn(NodalEquation *eqn) {
//   Trace("FEMesh::RemoveNodalEqn");
  int index = eqn->ndq_index();
  delete eqn;
  // Don't shift the rest of the vector entries to fill the hole
  // caused by deleting this eqn.  The hole can be filled more
  // efficiently by garbage collection later.  Just make sure to check
  // for null pointers when using the dof list!
  nodaleqn[index] = 0;
  nodaleqn_list_needs_cleaning=true;
}

void FEMesh::clean_nodaleqn() {
  if(nodaleqn_list_needs_cleaning) {
    std::vector<NodalEquation*>::size_type j = 0;
    for(std::vector<NodalEquation*>::size_type i=0; i<nodaleqn.size(); i++) {
      if(nodaleqn[i] != 0) {
	if(i != j) {
	  nodaleqn[j] = nodaleqn[i];
	  nodaleqn[j]->index_ = j;
	}
	j++;
      }
    }
    if(j != nodaleqn.size())
      nodaleqn.resize(j);
    nodaleqn_list_needs_cleaning = false;
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Make sure that all dofs and nodalequations have the right indices

void FEMesh::housekeeping() {
  clean_doflist();
  clean_nodaleqn();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

Node *FEMesh::newMapNode(const Coord &pos) {
  Node *node = new Node(ncount++, pos);
  mapnode.push_back(node);
  return node;
}

FuncNode *FEMesh::newFuncNode(const Coord &pos) {
  FuncNode *node = new FuncNode(this, ncount++, pos);
  funcnode.push_back(node);
  return node;
}

void FEMesh::reserveFuncNodes(int n) {
  funcnode.reserve(n);
}

void FEMesh::reserveMapNodes(int n) {
  mapnode.reserve(n);
}

void FEMesh::reserveElements(int n) {
  element.reserve(n);
}

int FEMesh::nnodes() const {
  // don't just return ncount here, since the node index is just an
  // identifier.  When nodes are deleted, ncount won't be decreased.
  return funcnode.size() + mapnode.size();
}

void FEMesh::AddElement(Element *el) {
  element.push_back(el);
  el->set_index(element.size()-1);
}

Element *FEMesh::getElement(int i) const {
  return element[i];
}

int FEMesh::nelements() const {
  return element.size();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

FEMesh::EquationData::EquationData()
  : active_index(-1),
    active(0),
    kin_active(0),
    dyn_active(0)
{}

void FEMesh::activate_equation(Equation &eqn) {
  // Trace("FEMesh::activate_equation " + eqn.name());
  int index = eqn.index();
  // Make sure that there's an EquationData entry for this Equation
  if(index >= int(eqndata.size())) {
    eqndata.resize(index+1);
  }
  if(!eqndata[index].active) {
    // Mark the Equation as active
    eqndata[index].active = true;
    // Offset of this Equation in the Nodes' NodalEquation lists
    eqndata[index].active_index = n_active_eqn;
    // Total number of active NodalEquations at each Node.
    n_active_eqn += eqn.dim();
    // Create NodalEquations for this Equation in each Node
    for(FuncNodeIterator ni(this); !ni.end(); ++ni) {
      (*this)[ni]->AddEquation(this, eqn);
    }
    // Activate fluxes required for this Equation
    eqn.activate_fluxes(this);
  }
}


void FEMesh::deactivate_equation(Equation &eqn) {
  if(is_active_equation(eqn)) {
    int index = eqn.index();
    eqndata[index].active = false;
    for(FuncNodeIterator ni(this); !ni.end(); ++ni) {
      (*this)[ni]->RemoveEquation(this, eqn);
    }
    // perform garbage collection on the nodaleqn list
    clean_nodaleqn();

    // update each active Equation's active_index
    n_active_eqn = 0;
    for(std::vector<EquationData>::size_type i=0; i<eqndata.size(); i++) {
      if(eqndata[i].active) {
	eqndata[i].active_index = n_active_eqn;
	n_active_eqn += Equation::alleqns()[i]->dim();
      }
    }
    eqn.deactivate_fluxes(this);
  }
}


bool FEMesh::is_active_equation(const Equation &eqn) const {
  std::vector<EquationData>::size_type index = eqn.index();
  return (index < eqndata.size()) && eqndata[index].active;
}


int FEMesh::active_equation_index(const Equation &eqn) const {
  // Offset of this Equation in the Nodes' NodalEquation lists
  return eqndata[eqn.index()].active_index;
}


int FEMesh::n_active_eqns() const {
  return n_active_eqn;
}

/*
  kinetic API starts here
*/


void FEMesh::kinetic_activate_equation(Equation &eqn) {
  if (eqn.get_kinetic() >0)
    eqndata[eqn.index()].kin_active = true;
  //    eqn.activate_kinetics();
}


void FEMesh::kinetic_deactivate_equation(Equation &eqn) {
  if(is_kinetically_active_equation(eqn) && eqn.get_kinetic() >0)
    //eqn.deactivate_kinetics();
    eqndata[eqn.index()].kin_active = true;
}


bool FEMesh::is_kinetically_active_equation(Equation &eqn) {
  return is_active_equation(eqn) && eqndata[eqn.index()].kin_active; //eqn.is_kinetically_active();
}

bool FEMesh::kinetics_activated() {
  for(std::vector<Equation*>::size_type i=0; i< all_equations()->size(); i++)
    {
      Equation * eqn = (*all_equations())[i];
      if (is_kinetically_active_equation(*eqn) == true) {
	return true;
      }
    }
  return false;
}


/*
  kinetic API ends here
*/

/*
  dynamics API starts here
*/


void FEMesh::activate_dynamics(Equation &eqn) {
  if (eqn.get_dynamics() >0)
    //eqn.activate_dynamics();
    eqndata[eqn.index()].dyn_active = true;
}


void FEMesh::deactivate_dynamics(Equation &eqn) {
  if(is_dynamically_active_equation(eqn) && eqn.get_dynamics() >0)
    //eqn.deactivate_dynamics();
     eqndata[eqn.index()].dyn_active = false;
}


bool FEMesh::is_dynamically_active_equation(Equation &eqn) {
  if(is_active_equation(eqn))
    if (eqndata[eqn.index()].dyn_active)
      {
	return true;
      }
  return false;
}

bool FEMesh::dynamics_activated() {
  for(std::vector<Equation*>::size_type i=0; i<all_equations()->size(); i++)
    {
      Equation * eqn  = (*all_equations())[i];
      if (is_dynamically_active_equation(*eqn) == true)
      return true;
    }
  return false;
}

/*
  dynamics API ends here
*/



//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void FEMesh::activate_flux(const Flux &flux) {
  //  std::cerr << "FEMesh::activate_flux: flux = " << flux.name() << std::endl;
  std::vector<int>::size_type index = flux.index();
  if(index >= active_flux.size()) {
    active_flux.resize(index+1, 0);
  }
  ++active_flux[index];
}

void FEMesh::deactivate_flux(const Flux &flux) {
  std::vector<int>::size_type index = flux.index();
  if(index < active_flux.size()) {
    if(active_flux[index] > 0)
      --active_flux[index];
  }
}

bool FEMesh::is_active_flux(const Flux &flux) const {
  std::vector<int>::size_type index = flux.index();
  if(index < active_flux.size()) {
    return active_flux[index] > 0;
  }
  return false;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// activate, etc, for Fields have to be routed through Field virtual
// functions, because they have to know if the Field is a
// CompoundField or not.  The routines that do the real work are
// FEMesh::do_activate, etc.

 class FieldMsg : public OOFMessage {
 public:
   FieldMsg(const std::string &msg, const FEMesh &mesh, const Field &field)
     : OOFMessage(msg)
   {
   //     Trace("FieldMsg: " + field.name() + " " + field.classname() + " " + to_string(&field));
     addarg(mesh);
     addarg(field.name());
   }
 };

Field &FEMesh::define_field(Field &field) {
  // Trace("FEMesh::define_field " + field.name());
  field.define(this);
  return field;
}

void FEMesh::undefine_field(Field &field) {
//   Trace("FEMesh::undefine_field " + field.name());
  field.undefine(this);
}

void FEMesh::activate_field(Field &field) {
  // Trace("FEMesh::activate_field " + field.name());
  field.activate(this);
}

void FEMesh::deactivate_field(Field &field) {
//   Trace("FEMesh::deactivate_field " + field.name());
  field.deactivate(this);
}

// Function to copy a field from another mesh.  Other mesh must have
// the same geometry, i.e. the same-indexed funcnodes must be in the
// same location, and the field must be defined.  Called by
// meshmenu.copyMesh.
void FEMesh::acquire_field_data(Field &field, const FEMesh *other) {
  for(FuncNodeIterator i(this); !i.end(); ++i) {
    for(int d=0; d<field.ndof(); ++d) {
      field(this, (*this)[i], d)->value() = \
	field(other, (*other)[i], d)->value();
    }
  }
}

FEMesh::FieldData::FieldData()
  : defined_index(-1),
    active(0),
    defined(0),
    in_plane(0)
{}

void FEMesh::do_define_field(Field &field) {
  // Trace("FEMesh::do_define_field " + field.name() + " " + to_string(field.index()));
  requirePrecompute();
  std::vector<FieldData>::size_type index = field.index();
  if(fielddata.size() <= index) {
    fielddata.resize(index+1);
  }
  FieldData &fdata = fielddata[index];
  if(!fdata.defined) {
    fdata.defined = true;
    fdata.defined_index = n_defined_dof;
    n_defined_dof += field.ndof();
    for(FuncNodeIterator ni(this); !ni.end(); ++ni) {
      (*this)[ni]->AddField(this, field);
    }
  }
}

int FEMesh::defined_index(const Field &field) const {
  return fielddata[field.index()].defined_index;
}

void FEMesh::do_undefine_field(Field &field) {
//   Trace("FEMesh::do_undefine_field " + field.name())
  requirePrecompute();
  if(is_defined_field(field)) {
    if(is_active_field(field)) {
      deactivate_field(field);
    }
    std::vector<FieldData>::size_type index = field.index();
    if(index < fielddata.size()) {
      fielddata[index].defined = false;
      for(FuncNodeIterator ni(this); !ni.end(); ++ni) {
	(*this)[ni]->RemoveField(this, field);
      }
      clean_doflist();

      // update each defined Field's defined_index
      n_defined_dof = 0;
      for(std::vector<FieldData>::size_type i=0; i<fielddata.size(); i++) {
	if(fielddata[i].defined) {
	  fielddata[i].defined_index = n_defined_dof;
	  n_defined_dof += Field::allfields()[i]->ndof();
	}
      }
    }
  }
}

bool FEMesh::is_defined_field(const Field &field) const {
//  std::cerr << "field.index()=" << field.index() << " fielddata=" << fielddata.size() << std::endl;
  return (field.index() < int(fielddata.size()))
    && fielddata[field.index()].defined;
}

void FEMesh::do_activate_field(Field &field) {

  if(!is_defined_field(field)) {
    define_field(field);	// allocates space in fielddata array
  }
  if(!is_active_field(field)) {
//     if(field.index() >= fielddata.size()) {
//       fielddata.resize(field.index()+1);
//     }
    n_active_field++;
    FieldData &fdata = fielddata[field.index()];
    fdata.active = true;
//     fdata.active_index = n_active_dof;
//     n_active_dof += field.ndof();
    // Set flags in degrees of freedom
    for(FuncNodeIterator ni(this); !ni.end(); ++ni) {
      for(int i=0; i<field.ndof(); i++) {
	field(this, (*this)[ni], i)->activate();
      }
    }
  }
}

void FEMesh::do_deactivate_field(Field &field) {
  if(is_active_field(field)) {
    n_active_field--;
    fielddata[field.index()].active = false;
//    // update each active Field's active_index
//     n_active_dof = 0;
//     for(int i=0; i<fielddata.size(); i++) {
//       if(fielddata[i].active) {
// 	fielddata[i].active_index = n_active_dof;
// 	n_active_dof += Field::allfields()[i]->ndof();
//       }
//     }
    // Unset flags in degrees of freedom
    for(FuncNodeIterator ni(this); !ni.end(); ++ni) {
      for(int i=0; i<field.ndof(); i++) {
	field(this, (*this)[ni], i)->deactivate();
      }
    }
  }
}

bool FEMesh::is_active_field(const Field &field) const {
  return (field.index() < int(fielddata.size()))
    && fielddata[field.index()].active;
}

int FEMesh::n_active_fields() const {
  return n_active_field;
}

void FEMesh::set_in_plane(const Field &field, bool planarity) {
  if(field.index() >= int(fielddata.size())) {
    fielddata.resize(field.index()+1);
  }
  fielddata[field.index()].in_plane = planarity;
}

bool FEMesh::in_plane(const Field &field) const {
  return (field.index() < int(fielddata.size()))
    && fielddata[field.index()].in_plane;
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Return a list of all the defined compound fields.  Compound fields
// are the ones that correspond directly to physical fields.
std::vector<CompoundField*>* FEMesh::all_fields() const {
  std::vector<CompoundField*>* flist = new std::vector<CompoundField*>;
  std::vector<CompoundField*> &allfields = CompoundField::allcompoundfields();
  flist->reserve(fielddata.size());
  for(std::vector<CompoundField*>::size_type i=0; i<allfields.size(); i++) {
    if(is_defined_field(*allfields[i])) 
      flist->push_back(allfields[i]);
  }
  return flist;
}

// Likewise for Equation objects.
std::vector<Equation*>* FEMesh::all_equations() const {
  std::vector<Equation*> *eqlist = new std::vector<Equation*>;
  std::vector<Equation*> &alleqns = Equation::alleqns();
  eqlist->reserve(eqndata.size());
  for(std::vector<Equation*>::size_type i=0; i<alleqns.size(); i++) {
    if(is_active_equation(*alleqns[i]))
      eqlist->push_back(alleqns[i]);
  }
  return eqlist;
}

std::vector<Flux*>* FEMesh::all_fluxes() const {
  std::vector<Flux*> *fluxlist = new std::vector<Flux*>;
  std::vector<Flux*> &allfluxen = Flux::allfluxes();
  for(std::vector<Flux*>::size_type i=0; i<allfluxen.size(); i++) {
    if(is_active_flux(*allfluxen[i]))
      fluxlist->push_back(allfluxen[i]);
  }
  return fluxlist;
}

std::vector<Flux*> FEMesh::allFluxes() const {  // non-swigged version.
  std::vector<Flux*> fluxlist;
  std::vector<Flux*> &allfluxen = Flux::allfluxes();
  for(std::vector<Flux*>::size_type i=0; i<allfluxen.size(); i++) {
    if(is_active_flux(*allfluxen[i]))
      fluxlist.push_back(allfluxen[i]);
  }
  return fluxlist;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// void FEMesh::precompute() {
//   // Things that have to be done before anything else can be calculated.
//   Material::precompute_all();
// }

void FEMesh::make_linear_system() {
  Trace("FEMesh::make_linear_system");
  Progress * localbar = getProgress();
  housekeeping();
  // std::cerr << "nodal equations =" << nodaleqn.size() << std::endl;
  // std::cerr << "dofs =" << dof.size() << std::endl;
  /*
    SPATIAL CONTRIBUTIONS START HERE
  */
  if(stiffness) delete stiffness;
  // Computational part of the right-hand-side.
  if(body_force_rhs) delete body_force_rhs;
  stiffness = new StiffnessMatrix(this, nodaleqn.size(), dof.size());
  body_force_rhs = new VECTOR_D(nodaleqn.size(), 0.0);  
  /*
    SPATIAL CONTRIBUTIONS END HERE
  */
  
  /*
    TIME-DEPENDENT CONTRIBUTIONS START HERE
  */
  if(c_matrix)
    {
      delete c_matrix;
    }
  if (kinetics_activated())
    {
      //returns true if ONE equation has kinetics enabled
      c_matrix = new StiffnessMatrix(this, nodaleqn.size(), dof.size());
      
    }
  else
    c_matrix = 0;
    
  if(m_matrix)
    {
      delete m_matrix;
    }
  if(dynamics_activated())
    //returns true if ONE equation has dynamics activated
    m_matrix = new StiffnessMatrix(this, nodaleqn.size(), dof.size());
  else
    m_matrix = 0;
    
  /*
    TIME-DEPENDENT CONTRIBUTIONS END HERE
  */

  std::string out_mssg;
  for(std::vector<Element*>::size_type i=0;
      i<element.size() && !localbar->query_stop(); i++)
    {
      if (solve_all_state() || element[i]->issubmesh())
	{
	  //	std::cerr << "adding element =" << i << std::endl;
	  element[i]->make_linear_system(this, stiffness, c_matrix, m_matrix, *body_force_rhs);
	  localbar->set_progress(float(i+1)/float(element.size()));
	  out_mssg = "creating linear system: " + to_string(i+1) + "/" + to_string(element.size());
	  localbar->set_message(out_mssg);
	}
    }
}


// Preliminary -- no housekeeping, just run the element's
// post-equilibration loop.
void FEMesh::post_process() {
  //Thread control starts here
  std::string out_mssg;
  Progress * localbar = getProgress(); //not a law firm
  //Thread control ends here
  for(std::vector<Element*>::size_type i=0;
      i< element.size() && !localbar->query_stop(); i++)
    {
      if (solve_all_state() || element[i]->issubmesh())
	{
	  element[i]->post_process(this);
	  localbar->set_progress(float(i+1)/float(element.size()));
	  out_mssg = "postprocessing: " + to_string(i+1) + "/" + to_string(element.size());
	  localbar->set_message(out_mssg);
	}
    }
}



const StiffnessMatrix &FEMesh::stiffness_matrix() const {
  if(!stiffness) {
    throw ErrProgrammingError("Stiffness matrix not computed!",
			      __FILE__, __LINE__);
  } 
  return *stiffness;
}

const StiffnessMatrix &FEMesh::conductivity_matrix() const {
  if(!c_matrix) {
    throw ErrProgrammingError("Conductivity matrix not computed!",
			      __FILE__, __LINE__);
  } 
  return *c_matrix;
}
const StiffnessMatrix &FEMesh::mass_matrix() const {
  if(!m_matrix) {
    throw ErrProgrammingError("Mass matrix not computed!",
			      __FILE__, __LINE__);
  } 
  return *m_matrix;
}

// Analogous to the above.
VECTOR_D &FEMesh::boundary_rhs() const {
  if(!boundary_right_side) {
    throw ErrProgrammingError("Boundary right hand side not available!",
			      __FILE__, __LINE__);
  }
  return *boundary_right_side;
}

void FEMesh::boundary_rhs_init(int n) {
  if (boundary_right_side) {
    delete boundary_right_side;
    // *boundary_right_side = 0.0; // Assigns to all of the elements.
  }
  boundary_right_side = new VECTOR_D(n, 0.0);
}

void FEMesh::dump_boundary_rhs() {
  std::cerr << "Boundary RHS:" << std::endl;
  for(VECTOR_D::size_type i=0; i<boundary_right_side->size(); i++) {
    std::cerr << i << " : " << (*boundary_right_side)[i] << std::endl;
  }
}

void FEMesh::refreshMaterials(PyObject *skeleton) {
  requirePrecompute();
  for(ElementIterator ei=element_iterator(); !ei.end(); ++ei)
    ei.element()->refreshMaterial(skeleton);
}

void FEMesh::requirePrecompute() {
  precomputeLock.acquire();
  precomputeRequired = true;
  precomputeLock.release();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

NodeIterator::NodeIterator(const FEMesh *const m, NodeIteratorType t)
  : mesh(m),
    type(t),
    index(0)
{}

NodeIterator FEMesh::node_iterator(NodeIteratorType t) const {
  return NodeIterator(this, t);
}

void NodeIterator::operator++() {
  if(int(index) != mesh->nnodes())
    index++;
}

bool NodeIterator::begin() const {
  return index == 0;
}

bool NodeIterator::end() const {
  return int(index) == mesh->nnodes();
}

// Caution: NodeIterator::index is not necessarily the same as
// node.index().  The argument to FEMesh::getNode is the
// NodeIterator::index.

Node *FEMesh::operator[](const NodeIterator &ni) const {
  return getNode(ni.index);
}

Node *NodeIterator::node() const {
  return mesh->getNode(index);
}


Node *FEMesh::getNode(int i) const {
  if(i < int(funcnode.size()))
    return funcnode[i];
  return mapnode[i - funcnode.size()];
}

// Finding the closest node to the mouse point.
// Used in MeshInfo
// TODO LATER: use hash table lookup here, instead of looping over all nodes.
Node *FEMesh::closestNode(const double x, const double y) {
  double min=1.;   // (initial value provided to suppress compiler warnings)
  Node *node, *thenode=0;
  for(NodeIterator ni = node_iterator(); !ni.end(); ++ni) {
    node = ni.node();
    double dx = node->position()(0) - x;
    double dy = node->position()(1) - y;
    double dist = dx*dx + dy*dy;
    if (ni.begin()) {
      min = dist;
      thenode = node;
    }
    else {
      if (dist <= min) {
	min = dist;
	thenode = node;
      }
    }
  }
  return thenode;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//


FuncNodeIterator::FuncNodeIterator(const FEMesh * const g, NodeIteratorType t)
  : mesh(g),
    type(t),
    index(0)
{}

FuncNodeIterator FEMesh::funcnode_iterator(NodeIteratorType t) const {
  return FuncNodeIterator(this, t);
}

void FuncNodeIterator::operator++() {
  if(index != mesh->funcnode.size()) index++;
}

bool FuncNodeIterator::end() const {
  return index == mesh->funcnode.size();
}

FuncNode *FEMesh::operator[](const FuncNodeIterator &ni) const {
  return funcnode[ni.index];
}

FuncNode *FuncNodeIterator::node() const {
  return mesh->funcnode[index];
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

ElementIterator::ElementIterator(const FEMesh * const msh, ElementIteratorType t)
  : mesh(msh),
    index(0),
    type(t)
{}

ElementIterator FEMesh::element_iterator(ElementIteratorType t) const {
  return ElementIterator(this, t);
}

void ElementIterator::operator++() {
  if(!end()) index++;
}

bool ElementIterator::end() const {
  return index == mesh->element.size();
}

Element *ElementIterator::element() const {
  return mesh->element[index];
}

int ElementIterator::size() const {
  return mesh->nelements();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Routines used to pick out rows and columns of the stiffness matrix
// when extracting submatrices.

bool FEMesh::solve_for_this_dof(int i) const {
  //dof list corresponds to unsymmetric system
  if (solve_all) //do the usual for the whole mesh
    return stiffness->is_nonempty_col(i) &&
      dof[i]->isfree() && dof[i]->active();
  //otherwise check that dof is part of selected submesh
  return stiffness->is_nonempty_col(i) &&
    dof[i]->isfree() && dof[i]->active() && dof[i]->issubmesh();
}

bool FEMesh::dont_solve_for_this_dof(int i) const {
  //dof list corresponds to unsymmetric system
  if (solve_all) //do the usual for the whole mesh
    return stiffness->is_nonempty_col(i) &&
      (!dof[i]->isfree() || !dof[i]->active());
  return stiffness->is_nonempty_col(i) &&
    (!dof[i]->isfree() || !dof[i]->active() || !dof[i]->issubmesh());
}

bool FEMesh::solve_this_nodal_eqn(int j) const {
  //nodaleqn corresponds to unsymmetric system
  NodalEquation *ndq = nodaleqn[j];
  bool result = stiffness->is_nonempty_row(j) && !ndq->is_dependent();
  if (solve_all) //do the usual for the whole mesh
    return result;
  return result && ndq->issubmesh();
  // return !nodaleqn[j]->is_dependent();
}

bool FEMesh::dont_solve_this_nodal_eqn(int j) const {
  if (solve_all) //do the usual for the whole mesh
    return stiffness->is_nonempty_row(j) && nodaleqn[j]->is_dependent();
  return stiffness->is_nonempty_row(j) &&
    (nodaleqn[j]->is_dependent() || !nodaleqn[j]->issubmesh());
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Extract the submatrices of the stiffness matrix
// The conceptual subdivision is this:
//
//                 free      fixed
//                          |
//   independent      A     | C
//                          |
//                ----------+---     --- 
//     dependent      C'    | B
// 

// Rowmapper and Colmapper.
// Build a map of the rows i of the stiffness matrix for
// which f(i) is true, and return it.  Call with f(i)
// discriminating between independent (known-force) vs.
// dependent (unknown-force) equations for rows, or
// with f(i) discriminating between fixed and free
// DOFs for columns.
//
// Do this by manipulating the pre-existing mapping list.
void FEMesh::rcmapper(bool (FEMesh::*f)(int) const,
		      std::vector<int>& rcmap, int len) const {
  //  Trace("FEMesh::rcmapper");
  //  vector<int> rowmap( nrows(), -1);
  // Double-check the size?
  int nrc = 0; // # of rows or columns in the mapped submatrix
  rcmap.resize(len); 
  for(int i=0; i<len; i++) {
    if ((this->*f)(i)) {
      rcmap[i] = nrc++;
    }
    else {
      rcmap[i] = -1; // -1 is default for rows/columns which don't occur.
    }
  }
}

// Run the maps, creating the lists used to partition the stiffness
// matrix.

void FEMesh::mapdofeqs() {
  Trace("FEMesh::mapdofeqs");
  rcmapper(&FEMesh::solve_this_nodal_eqn, indepeqnmap, stiffness->nrows());
  // cerr << "  indepeqnmap= "; print(indepeqnmap, cerr); cerr << endl;
  rcmapper(&FEMesh::dont_solve_this_nodal_eqn, depeqnmap, stiffness->nrows());
  // cerr << "    depeqnmap= "; print(depeqnmap, cerr); cerr << endl;
  rcmapper(&FEMesh::solve_for_this_dof, freedofmap, stiffness->ncols()); 
  // cerr << "   freedofmap= "; print(freedofmap, cerr); cerr << endl;
  rcmapper(&FEMesh::dont_solve_for_this_dof, fixeddofmap, stiffness->ncols()); 
  // cerr << "  fixeddofmap= "; print(fixeddofmap, cerr); cerr << endl;
} 


// Undo BC invocation by resetting dofs and nodalequations to be free
// and independent.  Does not modify "active" or "submesh" flags.
void FEMesh::resetdofeqs() {
  for(std::vector<DegreeOfFreedom*>::size_type i=0;i<dof.size();i++) 
    dof[i]->free();
  for(std::vector<NodalEquation*>::size_type i=0;i<nodaleqn.size();i++) {
    nodaleqn[i]->independent();
    // nodaleqn[i]->value() = 0.0;
  }
}


// Former routine from the FloatBC class, makes more sense here.
// This is a private routine that just does one map, called
// from FEMesh::cleanlists(). 
void FEMesh::cleanmap(std::vector<int>& rcmap) {
  std::vector<int> histogram(rcmap.size(),0);
  std::vector<int> subtractor(rcmap.size(),0);

  // Make the histogram.
  for(std::vector<int>::size_type i=0;i<rcmap.size();i++) {
    if (rcmap[i]!=-1)
      histogram[ rcmap[i] ]++;
  }
  // To actually clean the list, the technique is this:
  // Move along the histogram, counting the number of zero
  // entries you see.  When you get to a nonzero entry, 
  // make an entry in the subtraction map -- the amount
  // to subtract from the value you're looking at is the
  // current zero count.  Then, make a pass through the
  // subtraction map and do the actual subtractions.
  int zcount = 0;
  for(std::vector<int>::size_type i=0 ; i<rcmap.size(); i++) {
    if(histogram[i]==0) {
      zcount++;
    }
    else {
      subtractor[i] = zcount;
    }
  }

  // Now do the subtraction.
  for(std::vector<int>::size_type i=0; i<rcmap.size(); i++) {
    if( rcmap[i] != -1 ) {
      rcmap[i] -= subtractor[ rcmap[i] ];
    }
  }
  

}


// The cleanlists operation restores the row and column maps to 
// consistency after the application of a FloatBC.  The FloatBC's
// cause several rows or columns of the original stiffness matrix
// to be mapped (summed, actually) to a single row or column of
// the extracted 'A' matrix used in the actual linear-algebra solver.
// As a result, the range of the maps can develop holes -- cleanlists()
// closes those holes, ensuring that the range of the map is a 
// contiguous group of integers from 0 to the appropriate N.
void FEMesh::cleanlists() {
  cleanmap(indepeqnmap);
  cleanmap(depeqnmap);
  cleanmap(freedofmap);
  cleanmap(fixeddofmap);
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Extract the submatrices of the stiffness matrix
// Mapping arrays should be set up by the time you call this.

CSRmatrix FEMesh::Amatrix() const {
//   Trace("FEMesh::Amatrix");
  return stiffness->submatrix(indepeqnmap,freedofmap);
}

CSRmatrix FEMesh::Bmatrix() const {
//   Trace("FEMesh::Bmatrix");
  return stiffness->submatrix(depeqnmap,fixeddofmap);
}

CSRmatrix FEMesh::Cmatrix() const {
//   Trace("FEMesh::Cmatrix");
  return stiffness->submatrix(indepeqnmap,fixeddofmap);
}

CSRmatrix FEMesh::Cpmatrix() const {
//   Trace("FEMesh::Cpmatrix");
  return stiffness->submatrix(depeqnmap,freedofmap);
}

// Return the "blocking" data -- this is a map which is -1 if the
// corresponding row/column is in-plane, and is the node number if the
// corresponding row/column is out-of-plane.  Used by the block
// preconditioner, and by the Kinetic time stepping code.

// TODO: It might be more clever if these returned pointers/references
// to objects in either the femesh or stiffness matrix classes.
std::vector<int> FEMesh::A_rowblocks() const {
  return stiffness->rowgroups(indepeqnmap);
}

std::vector<int> FEMesh::A_colblocks() const {
  return stiffness->colgroups(freedofmap);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void FEMesh::get_unknowns(VECTOR_D &x, const std::vector<int> &map) const {
  for(std::vector<DegreeOfFreedom*>::size_type i=0; i<dof.size(); i++) {
    if(map[i] != -1) {
      x[map[i]] += dof[i]->value();
    }
  }
}

void FEMesh::set_unknowns(const VECTOR_D &x, const std::vector<int> &map) {
//   Trace("FEMesh::set_unknowns");
//   dump_dof();
  for(std::vector<DegreeOfFreedom*>::size_type i=0; i<dof.size(); i++) {
    if(map[i] != -1) {
//       std::cerr << "Setting dof[" << i << "] = " << x[map[i]] << std::endl;
      dof[i]->value() = x[map[i]];
    }
  }
}

void FEMesh::get_rhs(VECTOR_D &rhs, const std::vector<int> &map) const {
//   Trace("FEMesh::get_rhs");
  for(std::vector<NodalEquation*>::size_type i=0; i<nodaleqn.size(); i++) {
    if(map[i] != -1) {
      rhs[map[i]] += (*body_force_rhs)[i] + (*boundary_right_side)[i];
      // Merge nodal equation values with boundary RHS contribs.
    }
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void FEMesh::dump_dof() const {
  for(FuncNodeIterator i(this); !i.end(); ++i) {
    const FuncNode &node = *(*this)[i];
    std::cout << "node = " << node << std::endl;
    for(std::vector<Field*>::size_type f=0; f<Field::allfields().size(); f++) {
      const Field &field = *Field::allfields()[f];
      if(field.is_defined(this)) {
	for(int j=0; j<field.ndof(); j++) {
	  std::cout << "\t" << field.name() << "[" << j << "] dof="
	       << *field(this, node, j) << std::endl;
	}
      }
    }
    for(std::vector<Equation*>::size_type e=0;e<Equation::alleqns().size();e++){
      const Equation &eqn = *Equation::alleqns()[e];
      if(eqn.is_active(this)) {
	for(int j=0; j<eqn.dim(); j++) {
	  std::cout << "\t" << eqn.name() << "[" << j << "] nodaleqn="
	       << *eqn.nodaleqn(this, node, j) << std::endl;
	}
      }
    }
  }
}


void FEMesh::dump_eqn() const {
  for(FuncNodeIterator i(this); !i.end(); ++i) {
    const FuncNode &node = *(*this)[i];
    std::cout << "node = " << node << std::endl;
    for(std::vector<Equation*>::size_type e=0;e<Equation::alleqns().size();e++){
      const Equation &eqn = *Equation::alleqns()[e];
      if(eqn.is_active(this)) {
	for(int j=0; j<eqn.dim(); j++) {
	  std::cout << "\t" << eqn.name() << "[" << j << "] nodaleqn="
	       << *eqn.nodaleqn(this, node, j) << std::endl;
	}
      }
    }
  }
}

// In the case of inconsistent conjugacy data, build a trivial
// equation map.
void FEMesh::set_trivial_equation_mapping() {
  if (global_equation_map) {
    delete global_equation_map;
  }
  global_equation_map = new std::vector<int>(nodaleqn.size());
  for(unsigned int i=0;i<global_equation_map->size();++i) {
    (*global_equation_map)[i]=i;
  }
}

void FEMesh::set_equation_mapping(const std::vector<int>&rows,
 				 const std::vector<int>&cols)
{
  // Build a mapping array such that, when you traverse
  // FEMesh::nodaleqn[map[i]] for i from 0 to FEMesh::nodaleqn.size(),
  // the nodal equations come up in the same order as their conjugate
  // degrees of freedom do when you index that array (FEMesh::dof)
  // directly.  Store this map in the global_equation_map variable.
  // Using this ordering at stiffness-matrix-construction time causes
  // the resulting stiffness matrix to be as symmetric as possible.

  // Any equations for which conjugate DOFs do not exist come at the
  // tail of this list in no particular order.

  // For any given i, and in a given node's lists, the cols[i]-th
  // degree of freedom in the node is conjugate to the rows[i]-th
  // nodal equation in the node, with i ranging over the number of
  // active equations or degrees of freedom -- the number of *active*
  // eqns/dofs must match, but there may be inactive dofs.  There are
  // never inactive equations.
  
  // The local_equation_map array tells you the node-list index of the
  // conjugate DOF if you give it the node-list index of a nodal
  // equation, or -1 if it's not mapped, which can happen if there is
  // no conjugacy data.

  // The size of the local map is the number of equations per
  // funcnode, which is just the number of active equations, since all
  // nodes have the same equations, for now.
  std::vector<int> local_equation_map(n_active_eqn,-1);

  for(std::vector<int>::size_type i=0; i<rows.size(); ++i) {
    local_equation_map[rows[i]]=cols[i];
  }
  
  if (global_equation_map) {
    delete global_equation_map;
  }
  global_equation_map = new std::vector<int> (nodaleqn.size(), -1);
  
  // Make sure the lists are contiguous.
  housekeeping();

  // First build the "naive" global-forward-map array -- just map each
  // equation to the row index of its conjugate degree of freedom.

  // Can't use a FuncNodeIterator, it's const.
  int max_target_row = -1;
  for(std::vector<FuncNode*>::const_iterator fni = funcnode.begin();
      fni!=funcnode.end(); ++fni) {
    for(std::vector<NodalEquation*>::size_type i=0; 
	i<((*fni)->eqnlist).size(); ++i) {
      // Only do stuff if this mapping makes sense.
      if (local_equation_map[i]!=-1) {
	NodalEquation* neq = ((*fni)->eqnlist)[i];
	int source_row = neq->ndq_index(); // Index in master stiffness mtx

	DegreeOfFreedom *dof = ((*fni)->doflist)[local_equation_map[i]];
	int target_row = dof->dofindex();  // Also in master stiffness mtx
	
	(*global_equation_map)[source_row] = target_row;
	max_target_row = (target_row > max_target_row ? target_row : 
			  max_target_row);
      }
    }
  }

  
  // There may be unmapped equations due to incomplete conjugacy data.
  // Also, there may be equations for which there are no degrees of
  // freedom because they're going to be dependent equations.  In
  // either case, these "extra" equations will not have been mapped
  // anywhere.  Make a second pass through the list, and assign any
  // unassigned equations an incremental index, starting with
  // max_target_row.  Bounds are not an issue, because of the
  // range-compression step which follows this block.
  for(std::vector<FuncNode*>::const_iterator fni = funcnode.begin();
      fni!=funcnode.end(); ++fni) {
    for(std::vector<NodalEquation*>::size_type i=0; 
	i<((*fni)->eqnlist).size(); ++i) {
      NodalEquation *neq = ((*fni)->eqnlist)[i];
      int source_index = neq->ndq_index(); 
      if ( (*global_equation_map)[source_index] == -1 ) {
	max_target_row++;
	(*global_equation_map)[source_index]=max_target_row;
      }
    }
  }

  // Now fix the range of the global_equation_map -- the master
  // stiffness matrix can be wider than it is tall, so some of the
  // equations could be out of range.  Fix this by compressing the
  // range but preserving the order.

  // First, build a map indexed by the *range* elements of the map,
  // but preserving the domain information.  This is the "inverse" of
  // the forward-map.
  std::map<int,int> back_map;
  for(std::vector<int>::size_type i=0; i<global_equation_map->size(); ++i) {
    back_map[(*global_equation_map)[i]]=i;
  }
  // Then go through the backward map in range-of-forward-map
  // (domain-of-backward-map, key of std::map structure) order, and
  // for each element of the domain of the forward map (range of the
  // backward map, value of std::map structure), assign consecutive
  // integers to the range of the forward map.
  int key_count=0;
  for(std::map<int,int>::const_iterator bmi=back_map.begin();
      bmi!=back_map.end(); ++bmi) {
    (*global_equation_map)[(*bmi).second]=key_count;
    key_count++;
  }

}


// void FEMesh::set_local_mapping(const std::vector<int>&rows, 
// 		       const std::vector<int>&cols)
// {
//   //clear_symmetricity_maps();
//   /*
//     this function establishes relations between rows and columns,
//     so the effective stiffness matrix can be symmetrized
//   */
//   if(local_rows_mapping)
//     {
//       delete local_rows_mapping;  
//     }
//   local_rows_mapping = new std::vector<int> (rows.size());
  
  
//   // construct mapping to relate the local nodal equations to the 
//   // the to-solve-for degrees of freedom
//   for(int irow = 0; irow< rows.size(); irow++)
//     {
//       (*local_rows_mapping)[rows[irow]] = cols[irow];
//       //std::cerr << irow <<"=" << cols[irow] << std::endl; 
//     }
//   //allocate memory for the forward and backward mapping
//   int global_mapping_size = nodaleqn.size(); //number of rows
//   if (forward_mapping)
//     {
//       delete forward_mapping;
      
//     }
//   forward_mapping = new std::vector<int> (global_mapping_size, -1);
// }

void FEMesh::clear_symmetricity_maps()
{
  //destroy the symmetricity maps after using them
  delete global_equation_map;
  global_equation_map=0;
}

// Having remapped the equation array, we need a helper function so
// that boundary conditions can set the appropriate equations to be
// dependent.  
// const NodalEquation & FEMesh::map_forward_local(const Equation & equation, 
// 			const FuncNode &node, const int &eqcomp)
// {

//   int local_target_row = (*local_equation_map)[equation.localindex(this, 
// 								   eqcomp)];
//   return *node.eqnlist[local_target_row];
// }

const NodalEquation & FEMesh::map_forward_equation(const Equation &equation,
						   const FuncNode &node,
						   const int &eqcomp) const {
  int target_index = map_forward_index(equation, node, eqcomp);
  return *(nodaleqn[target_index]);
}

int FEMesh::map_forward_index(const Equation & equation, 
			const FuncNode &node, const int &eqcomp) const 
{
  // Reads the global equation map -- see set_equation_mapping for details.
  int original_row = equation.nodaleqn(this, node, eqcomp)->ndq_index();
  return (*global_equation_map)[original_row];
}


int FEMesh::check_symmetricity(const CSRmatrix &A)
{
  for (int i=0; i<A.nrows(); i++)
    for (int j=0; j<A.ncols(); j++)
      if (A(i,j) != A(j,i))
	{
	  std::cerr << "(i,j)= (" << i <<"," << j << ")"<< std::endl;
	  std::cerr << "A(i,j)=" << A(i,j) << " A(j,i)=" << A(j,i) << std::endl;
	  return 0;
	}
  return 1;

}

CSRmatrix FEMesh::ConductivityMatrix() const {
  Trace("FEMesh::ConductivityMatrix");
  return c_matrix->submatrix(indepeqnmap,freedofmap);
}

CSRmatrix FEMesh::MassMatrix() const {
  Trace("FEMesh::MassMatrix");
  return m_matrix->submatrix(indepeqnmap,freedofmap);
}

//################//###################//################

// initializing "scpatches"
void FEMesh::init_scpatches(const std::vector<int> *snodes)
{
  for(std::vector<int>::size_type i=0; i<snodes->size(); i++)
    scpatches[(*snodes)[i]] = new NodalSCPatches();
}

// create/add a new CSCPatch pointer
void FEMesh::add_scpatch(const int assembly_node, const Material *mat,
			 const int order,
			 const std::vector<int> *elems,
			 const std::vector<int> *nodes,
			 const int qualified)
{
  // make sure to delete new'd CSCPatch's at FEMesh's destruction.
  CSCPatch *patch = new CSCPatch(this, order, mat, elems, nodes, qualified);
  scpatches[assembly_node]->add_patch(patch);
}

// getting elements and nodes from the patch
std::vector<int> *FEMesh::get_elements_from_patch(const int assembly_node,
						  const Material *mat)
{
  return scpatches[assembly_node]->get_elements_from_patch(mat);
}

std::vector<int> *FEMesh::get_nodes_from_patch(const int assembly_node,
					       const Material *mat)
{
  return scpatches[assembly_node]->get_nodes_from_patch(mat);
}

// initializing "recovered_fluxes"
void FEMesh::init_nodalfluxes()
{
  for(std::vector<FuncNode*>::size_type i=0; i<funcnode.size(); i++)
    recovered_fluxes[funcnode[i]->index()] = new NodalFluxes();
}

// recover fluxes
void FEMesh::recover_fluxes()
{
  // iterate over NodalSCPatches objects
  std::map<const int, NodalSCPatches*>::iterator iter;
  std::vector<Flux*> allfluxes = allFluxes();
  for(iter=scpatches.begin(); iter!=scpatches.end(); iter++)
    (iter->second)->recover_fluxes(allfluxes);
}

// adding recovered flux to the map
void FEMesh::add_this_flux(const Material *mat, const Flux *fluks,
			   const Node *nd, VECTOR_D *fvalues)
{
  recovered_fluxes[nd->index()]->add_flux_value(mat, fluks, fvalues);
}

// computing a recovered flux at MasterCoord
VECTOR_D *FEMesh::get_recovered_flux(const Flux *fluks,
				     const Element *elem,
				     const MasterCoord &mc)
{
  int dim = fluks->ndof();
  VECTOR_D *result = new VECTOR_D(dim, 0.0);
  for(int i=0; i<dim; i++) {
    // interpolating the value
    for(ElementFuncNodeIterator node = elem->funcnode_iterator(); !node.end(); ++node) {
      double temp = recovered_fluxes[(node.node())->index()]->get_flux_component(elem->material(), fluks, i);
      (*result)(i) += node.shapefunction(mc)*temp;
    }
  }
  return result;
}


// computing all the recovered fluxes at Coord -- swigged one
void FEMesh::report_recovered_fluxes(const Element *elem,
				  const Coord *point)
{
  std::vector<Flux*> fluxes = allFluxes();
  for (std::vector<Flux*>::size_type i=0; i<fluxes.size(); i++) {
    std::cerr << fluxes[i]->name() << std::endl;
    int dim = fluxes[i]->ndof();
    for(int j=0; j<dim; j++) {
      double j_th_component = 0.0;
      for(ElementFuncNodeIterator node = elem->funcnode_iterator(); !node.end(); ++node) {
	double comp = recovered_fluxes[(node.node())->index()]->get_flux_component(elem->material(), fluxes[i], j);
	j_th_component += node.shapefunction(elem->to_master(*point))*comp;
      }
      std::cerr << j_th_component << std::endl;
    }
    std::cerr << "==========================" << std::endl;
  }
}


double FEMesh::zz_L2_estimate(const Element *elem, const Flux *fluks)
{
  // What to compute:
  // Integral_e[(flux_ex-flux_fe)^T (flux_ex-flux_fe)]^(1/2)
  // flux_ex -> recovered flux
  // 
  // Things to consider to determine integration order
  // Linear mesh: recovered stress field is linear, so the order
  // is 2 -- 2x2 gauss point needed.
  // Quadratic mesh: recovered stress field is quadratic, thus the order
  // should be 4 -- 3x3 gauss points needed.

  int order = 2*elem->shapefun_degree();
  int dim = fluks->ndof();
  double error = 0.0;
  double refer = 0.0;
  for(GaussPointIterator gpt = elem->integrator(order); !gpt.end(); ++gpt) {
    MasterCoord mc = gpt.gausspoint().mastercoord();
    double wt = gpt.gausspoint().weight();
    zz_L2_estimate_sub(elem, fluks, dim, error, refer, mc, wt);
  }
  return sqrt(error/refer);
  // TODO: What if refer == 0?
}


void FEMesh::zz_L2_estimate_sub(const Element *elem, const Flux *fluks,
				const int &dim,
				double &error, double &refer,
				const MasterCoord &mc, const double &wt)
{
  // flux(diff) vectors
  VECTOR_D *recovered = get_recovered_flux(fluks, elem, mc);
  VECTOR_D *feflux = fluks->evaluate(this, elem, mc);
  VECTOR_D diff = *recovered - *feflux;
  // dot product while integrating
  for(int i=0; i<dim; i++) {
    error += diff(i)*diff(i);
    refer += (*recovered)(i) * (*recovered)(i);
  }
  error *= wt;
  refer *= wt;
  delete recovered;
  delete feflux;
}


std::vector<double> *FEMesh::zz_L2_weights(const Flux *fluks,
					   const double &bottom,
					   const double &top)
{
  std::vector<double> *weights = new std::vector<double>;
  std::vector<double> values;
  int order = 2*element[0]->shapefun_degree();
  int dim = fluks->ndof();

  Element *elem = element[0];
  double value = 0.0;
  for(GaussPointIterator gpt = elem->integrator(order); !gpt.end(); ++gpt) {
    MasterCoord mc = gpt.gausspoint().mastercoord();
    double wt = gpt.gausspoint().weight();
    zz_L2_weights_sub(elem, fluks, dim, value, mc, wt);
  }
  values.push_back(value);
  double min = value;
  double max = value;

  for(int i=1; i<nelements(); i++) {
    Element *elem = element[i];
    value = 0.0;
    for(GaussPointIterator gpt = elem->integrator(order); !gpt.end(); ++gpt) {
      MasterCoord mc = gpt.gausspoint().mastercoord();
      double wt = gpt.gausspoint().weight();
      zz_L2_weights_sub(elem, fluks, dim, value, mc, wt);
    }
    values.push_back(value);
    if(value < min)
      min = value;
    if(value > max)
      max = value;
  }

  double upper = min + (max-min)*top;
  double lower = min + (max-min)*bottom;
  double diff = upper - lower;
  double sum = upper + lower;
  for(std::vector<double>::size_type i=0; i<values.size(); i++)
    weights->push_back(0.5*erf(M_E*(2*values[i]-sum)/diff)+0.5);
  return weights;
}


void FEMesh::zz_L2_weights_sub(const Element *elem, const Flux *fluks,
			       const int &dim, double &value,
			       const MasterCoord &mc, const double &wt)
{
  // flux vector
  VECTOR_D *feflux = fluks->evaluate(this, elem, mc);
  // dot product while integrating
  for(int i=0; i<dim; i++) {
    value += (*feflux)(i)*(*feflux)(i);
  }
  value *= wt;
  delete feflux;
}



