// -*- C++ -*-
// $RCSfile: node.C,v $
// $Revision: 1.19 $
// $Author: reida $
// $Date: 2006/01/25 21:59:51 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include "common/tostring.h"
#include "common/trace.h"
#include "element.h"
#include "equation.h"
#include "femesh.h"
#include "nodalequation.h"
#include "node.h"

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Names needed for PythonExportable base class
const std::string Node::modulename_("oof2.SWIG.engine.node");
const std::string Node::classname_("Node");
const std::string FuncNode::classname_("FuncNode");

Node::Node(int n, const Coord &p)
  : pos(p),
    index_(n)
{}

FuncNode::FuncNode(FEMesh *mesh, int n, const Coord &p)
  : Node(n, p), doflist(0), eqnlist(0)
{
  // No space is allocated for dofs or nodaleqns here.  It's all done
  // by AddField and AddEquation.  That means that we're assuming that
  // the FEMesh construction process creates all of the nodes *before*
  // defining any fields or activating any equations.  
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Node comparison operator -- the node integer index 
// is the one true arbiter of node equality.

bool operator==(const Node &n1, const Node &n2)
{
  return n1.index() == n2.index();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void FuncNode::AddField(FEMesh *mesh, const Field &field) {
//   Trace("FuncNode::AddField " + field.name() + " " + to_string(index()) 
// 	+ " Pos: " + to_string(pos));
  doflist.reserve(doflist.size() + field.ndof());
  for(int i=0; i<field.ndof(); i++) {
    DegreeOfFreedom *dof = mesh->CreateDoF();
//     Trace("FuncNode::AddField adding DOF with index " + 
// 	  to_string(dof->dofindex()));
    doflist.push_back(dof);
    // doflist.push_back(mesh->CreateDoF());
  }
}

void FuncNode::RemoveField(FEMesh *mesh, const Field &field) {
  // Remove degrees of freedom from the doflist.
  // Use the messy STL iterator instead of int so that erase() can be used.
//   Trace("FuncNode::RemoveField " + field.name() + " " + to_string(index())
// 	+ " Pos: " + to_string(pos));
  std::vector<DegreeOfFreedom*>::iterator start =
    doflist.begin() + mesh->defined_index(field);
  std::vector<DegreeOfFreedom*>::iterator end = start + field.ndof();
  for(std::vector<DegreeOfFreedom*>::iterator i=start; i!=end; ++i) {
//     Trace("FuncNode::RemoveField removing DOF with index " + 
// 	  to_string( (*i)->dofindex() ));
    mesh->RemoveDoF(*i);	// remove them from the mesh
  }
  doflist.erase(start, end);	// and remove them from the node
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void FuncNode::AddEquation(FEMesh *mesh, const Equation &eqn) {
//   Trace("FuncNode::AddEquation " + eqn.name() + " " + to_string(index())
// 	+ " Pos: " + to_string(pos));
  eqnlist.reserve(eqnlist.size() + eqn.dim());
  for(int i=0; i<eqn.dim(); i++) {
    NodalEquation *ne = mesh->CreateNodalEqn();
//     Trace("FuncNode::AddEquation adding NE with index " + 
// 	  to_string(ne->index()));
    eqnlist.push_back(ne);
    // eqnlist.push_back(mesh->CreateNodalEqn());
  }
}

void FuncNode::RemoveEquation(FEMesh *mesh, const Equation &eqn) {
//   Trace("FuncNode::RemoveEquation " + eqn.name() + " " + to_string(index())
// 	+ " Pos: " + to_string(pos));
  std::vector<NodalEquation*>::iterator start =
    eqnlist.begin() + mesh->active_equation_index(eqn);
  std::vector<NodalEquation*>::iterator end = start + eqn.dim();
  for(std::vector<NodalEquation*>::iterator i=start; i!=end; ++i) {
//     Trace("FuncNode::RemoveEquation removing NE with index " + 
// 	  to_string( (*i)->index() ));
    mesh->RemoveNodalEqn(*i);
  }
  eqnlist.erase(start, end);
}

//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//

Coord FuncNode::displaced_position(const FEMesh *mesh) const {
  static TwoVectorField *displcmnt = 0;
  if(!displcmnt)
    displcmnt = dynamic_cast<TwoVectorField*>(Field::getField("Displacement"));
  if(displcmnt->is_defined(mesh)) {
    return pos + (*displcmnt)(mesh, this);
  }
  return pos;
}

//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//

// Representational stuff -- mostly for debugging

std::string Node::repr() const {
  std::string str = "Node(";
  str += to_string(position()) + ")";
  return str;
}

std::string FuncNode::repr() const {
  std::string str = "FuncNode(";
  str += to_string (position()) + ")";
  return str;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Print a string suitable for use in a Python constructor

std::string Node::ctor() const {
  return "newMapNode(Coord" + to_string(position()) + ")";
}

std::string FuncNode::ctor() const {
  return "newFuncNode(Coord" + to_string(position()) + ")";
}

//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//

// Boundary condition flags  -- LEFTOVER FROM OOF1 -- probably not needed

bool FuncNode::fixed() const {
  // Is any dof fixed?
  for(std::vector<DegreeOfFreedom*>::size_type i=0; i<doflist.size(); i++)
    if(!doflist[i]->isfree())
      return 1;
  return 0;
}


//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//


// LEFTOVER FROM OOF1

int Node::hashfunc(const Coord &where, double &dist) {
  Coord r(where - position());
  dist = r(0)*r(0) + r(1)*r(1);
  return 1;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

std::ostream &operator<<(std::ostream &os, const Node &node) {
  os << "[" << node.index() << "] " << node.pos;
  return os;
}
