// -*- C++ -*-
// $RCSfile: preconditioner.h,v $
// $Revision: 1.25 $
// $Author: reida $
// $Date: 2006/02/27 15:27:58 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#ifndef PRECONDITIONER_H
#define PRECONDITIONER_H

#include "csrmatrix.h"
#include "engine/matvec.h"
#include "SparseLib++/ilupre_double.h"
#include "SparseLib++/icpre_double.h"
#include "SparseLib++/diagpre_double.h"
#include "blockpre_double.h"

#include "common/pythonexportable.h"
#include <string>

#ifdef HAVE_PETSC
// #include "petscvec.h"
// #include "petscmat.h"
// #include "petscsles.h"
#endif //HAVE_PETSC

class PreconditionerCore {
public:
  virtual ~PreconditionerCore() {}
  virtual VECTOR_D solve(const VECTOR_D&) const = 0;
  virtual VECTOR_D trans_solve(const VECTOR_D&) const = 0;
#ifdef HAVE_PETSC
  // virtual void set(PC &precond) const = 0;
#endif //HAVE_PETSC
};

class Preconditioner;
class UnitPreconditioner;


class ILU_Preconditioner : public PreconditionerCore,
			   public CompRow_ILUPreconditioner_double
{
public:
  template <class MTX>
  ILU_Preconditioner(const MTX &A)
    : CompRow_ILUPreconditioner_double(A)
  {}
  virtual VECTOR_D solve(const VECTOR_D &x) const {
    return CompRow_ILUPreconditioner_double::solve(x);
  }
  virtual VECTOR_D trans_solve(const VECTOR_D &x) const {
    return CompRow_ILUPreconditioner_double::trans_solve(x);
  }
#ifdef HAVE_PETSC
  // virtual void set(PC &precond) const {};
#endif //HAVE_PETSC

};


class IC_Preconditioner : public PreconditionerCore,
			  public ICPreconditioner_double
{
public:
  template <class MTX>
  IC_Preconditioner(const MTX &A)
    : ICPreconditioner_double(A) {}
  virtual VECTOR_D solve(const VECTOR_D &x) const {
    return ICPreconditioner_double::solve(x);
  }
  virtual VECTOR_D trans_solve(const VECTOR_D &x) const {
    return ICPreconditioner_double::trans_solve(x);
  }
#ifdef HAVE_PETSC
  // virtual void set(PC &precond) const {};
#endif //HAVE_PETSC
};


class Un_conditioner : public PreconditionerCore {
public:
  template <class MTX>
  Un_conditioner(const MTX&) {}
  virtual VECTOR_D solve(const VECTOR_D &x) const {
    return x;
  }
  virtual VECTOR_D trans_solve(const VECTOR_D &x) const {
    return x;
  }
#ifdef HAVE_PETSC
  // virtual void set(PC &precond) const {};
#endif //HAVE_PETSC
};


class Jacobi_Preconditioner: public PreconditionerCore,
			     public DiagPreconditioner_double
{
public:
  template <class MTX>
  Jacobi_Preconditioner(const MTX &A)
    : DiagPreconditioner_double(A)
  {}
  virtual VECTOR_D solve(const VECTOR_D &x) const {
    return DiagPreconditioner_double::solve(x);
  }
  virtual VECTOR_D trans_solve(const VECTOR_D &x) const {
    return DiagPreconditioner_double::trans_solve(x);
  }
#ifdef HAVE_PETSC
  // virtual void set(PC &precond) const {};
#endif // HAVE_PETSC
};


class Block_Preconditioner: public PreconditionerCore,
			    public CompRow_BlockPreconditioner_double 
{
public:
  template <class MTX>
  Block_Preconditioner(const MTX &A, 
		       const std::vector<int> &rg,
		       const std::vector<int> &cg,
		       UnitPreconditioner *inplane)
    : CompRow_BlockPreconditioner_double(A, rg, cg, inplane)
  {}
  virtual VECTOR_D solve(const VECTOR_D &x) const {
    return CompRow_BlockPreconditioner_double::solve(x);
  }
  virtual VECTOR_D trans_solve(const VECTOR_D &x) const {
    return CompRow_BlockPreconditioner_double::trans_solve(x);
  }
#ifdef HAVE_PETSC
  // virtual void set(PC &precond) const {};
#endif // HAVE_PETSC
};




#ifdef HAVE_PETSC
// #include "petscvec.h"
// #include "petscmat.h"
// #include "petscsles.h"
// #include "engine/preconditioner.h"
#endif //HAVE_PETSC

class Preconditioner : public PythonExportable<Preconditioner> {
public:
  Preconditioner() {}
  virtual ~Preconditioner() {}
  virtual PreconditionerCore * create_preconditioner(const CSRmatrix &A,
						 const std::vector<int> &rg,
						 const std::vector<int> &cg)=0;
  static const std::string modulename_;
  virtual const std::string &modulename() const { return modulename_; }
#ifdef HAVE_PETSC
  // virtual void set(PC) const {};//sets correct preconditioner in derived class 
#endif //HAVE_PETSC
};


// Unit preconditioners are non-composite.  They are required to
// implement a simple construction function that only takes the
// matrix, in addition to the generic interface.  This simpler
// function is called by the composite preconditioner.
class UnitPreconditioner : public Preconditioner {
public:
  // build_preconditioner is the "local" preconditioner construction
  // call, not part of the exterior interface.
  virtual PreconditionerCore * build_preconditioner(const CSRmatrix &A)=0;
#ifdef HAVE_PETSC
  // virtual void set(PC) const {};
#endif // HAVE_PETSC
};


class ILUPreconditioner : public UnitPreconditioner { 
//wrapper for core class
public:
  ILUPreconditioner() {}
  virtual ~ILUPreconditioner(); 
  virtual PreconditionerCore * create_preconditioner(const CSRmatrix &A,
						 const std::vector<int> &rg,
						 const std::vector<int> &cg);
  virtual PreconditionerCore * build_preconditioner(const CSRmatrix &A);
  
  static const std::string classname_;
  virtual const std::string &classname() const { return classname_; }
  
};

class ICPreconditioner : public UnitPreconditioner  { 
//wrapper for core class
public:
  ICPreconditioner() {}
  virtual ~ICPreconditioner();
  virtual PreconditionerCore * create_preconditioner(const CSRmatrix &A,
						 const std::vector<int> &rg,
						 const std::vector<int> &cg);
  virtual PreconditionerCore * build_preconditioner(const CSRmatrix &A);
  static const std::string classname_;
  virtual const std::string &classname() const { return classname_; }  
};
    
class UnPreconditioner  : public UnitPreconditioner { 
//wrapper for core class
public:
  UnPreconditioner() {}
  virtual ~UnPreconditioner();
  virtual PreconditionerCore * create_preconditioner(const CSRmatrix &A,
						 const std::vector<int> &rg,
						 const std::vector<int> &cg);
  virtual PreconditionerCore * build_preconditioner(const CSRmatrix &A);
  static const std::string classname_;
  virtual const std::string &classname() const { return classname_; }  
}; 


class JacobiPreconditioner : public UnitPreconditioner {
public:
  JacobiPreconditioner() {}
  virtual ~JacobiPreconditioner();
  virtual PreconditionerCore * create_preconditioner(const CSRmatrix &A,
						 const std::vector<int> &rg,
						 const std::vector<int> &cg);
  virtual PreconditionerCore * build_preconditioner(const CSRmatrix &A);
  static const std::string classname_;
  virtual const std::string &classname() const { return classname_; }
};


// The Block preconditioner is composite, so it's not a subclass of
// UnitPreconditioner.
class BlockPreconditioner : public Preconditioner {
private:
  UnitPreconditioner *inplane;
public:
  BlockPreconditioner(UnitPreconditioner *ip) : inplane(ip) {}
  virtual ~BlockPreconditioner();

  virtual PreconditionerCore *create_preconditioner(const CSRmatrix &A,
						const std::vector<int> &rg, 
						const std::vector<int> &cg);
  static const std::string classname_;
  virtual const std::string &classname() const { return classname_; }
};

#endif // PRECONDITIONER_H
