// -*- C++ -*-
// $RCSfile: thermalexpansion.C,v $
// $Revision: 1.50 $
// $Author: langer $
// $Date: 2006/02/14 22:12:48 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>


#include "common/coord.h"
#include "engine/IO/propertyoutput.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/equation.h"
#include "engine/femesh.h"
#include "engine/flux.h"
#include "engine/material.h"
#include "engine/matvec.h"
#include "engine/property/elasticity/cijkl.h"
#include "engine/property/elasticity/elasticity.h"
#include "engine/IO/propertyoutput.h"
#include "engine/cstrain.h"
#include "engine/property/orientation/orientation.h"
#include "thermalexpansion.h"


ThermalExpansion::ThermalExpansion(PyObject *reg, 
				   const std::string &nm, 
				   double t0)
  : Property(nm,reg),
    tzero_(t0)
{
  temperature=dynamic_cast<ScalarField*>(Field::getField("Temperature"));
  stress_flux=dynamic_cast<SymmetricTensorFlux*>(Flux::getFlux("Stress"));
}


void ThermalExpansion::precompute() {

}

int ThermalExpansion::integration_order(const FEMesh*, 
					const Element *el) const {
  return el->shapefun_degree();
}


void ThermalExpansion::fluxmatrix(const FEMesh *mesh,
				  const Element *element,
				  const ElementFuncNodeIterator &nu, 
				  const Flux *flux, FluxData *fluxdata,
				  const MasterPosition &x) const
{
  // The stress-free strain is $\epsilon_{kl}^0(T) = \alpha T$
  // where alpha is a symmetric tensor.
  //
  // Our goal is to subtract C_{ijkl}*u_{kl}^0(T) from the stress
  // tensor.
  if(*flux != *stress_flux) {
    throw ErrProgrammingError("Unexpected flux." __FILE__, __LINE__);
  }
  
  // Assume linear elasticity, for now.  The "elasticity" variable was
  // set by cross_reference().
  const Cijkl modulus = elasticity->cijkl(mesh, element, x);

  double sfval = nu.shapefunction(x);
  // Loop over all tensor components.
  SymTensorIterator kl;		// reuse inner loop iterator
  for(SymTensorIterator ij; !ij.end(); ++ij) {
    double &mtx_el = fluxdata->matrix_element(mesh, ij, temperature, nu); 
    // Loop over strain tensor components.
    for( ; !kl.end(); ++kl) {
      double ca = modulus(ij, kl)*expansiontensor(kl.row(), kl.col())*sfval;
      if(kl.diagonal()) {
	mtx_el -= ca;
      }
      else {
	mtx_el -= 2.0*ca;
      }
    }
    kl.reset();
  }
}

void ThermalExpansion::fluxrhs(const FEMesh *mesh, const Element *element,
				   const Flux *flux, FluxData *fluxdata,
				   const MasterPosition &x) const {
  
  if(*flux!=*stress_flux) {
    throw ErrProgrammingError("Unexpected flux." __FILE__, __LINE__);
  }
  const Cijkl modulus = elasticity->cijkl(mesh, element, x);

  for(SymTensorIterator ij; !ij.end(); ++ij) {
    double &offset_el = fluxdata->offset_element(mesh, ij); // reference!
    for(SymTensorIterator kl; !kl.end(); ++kl) {
      if(kl.diagonal()) {
	offset_el += modulus(ij,kl)*expansiontensor[kl]*tzero_;
      }
      else {
	offset_el += 2.0*modulus(ij,kl)*expansiontensor[kl]*tzero_;
      }
    }
  }
}


void ThermalExpansion::output(const FEMesh *mesh,
				  const Element *element,
				  const PropertyOutput *output,
				  const MasterPosition &pos,
				  OutputVal *data)
  const 
{
  const std::string &outputname = output->name();
  if(outputname == "Strain") {
    // The parameter is a Python StrainType instance.  Extract its name.
    std::string stype = output->getRegisteredParamName("type");
    SymmMatrix3 *sdata = dynamic_cast<SymmMatrix3*>(data);
    // Compute alpha*T with T interpolated to position pos
    // TODO LATER: This is baroque and may be slow.  Make a
    // specialized function for getting the value of a scalar field.
    const OutputValue tfield = element->outputField(mesh, *temperature, pos);
    const ScalarOutputVal *tval = 
      dynamic_cast<const ScalarOutputVal*>(tfield.valuePtr());
    double t = tval->value();
    

    if(stype == "Thermal")
      *sdata += expansiontensor*(t-tzero_);
    else if(stype == "Elastic")
      *sdata -= expansiontensor*(t-tzero_);
  }//strain output ends here
  if(outputname == "Energy") {
    std::string etype = output->getEnumParam("etype");
    if(etype == "Total" || etype == "Elastic") {
      ScalarOutputVal *edata = 
	dynamic_cast<ScalarOutputVal*>(data);
      SymmMatrix3 thermalstrain;
      const Cijkl modulus = elasticity->cijkl(mesh, element, pos); //returns lab reference system stiffness
      const OutputValue tfield = element->outputField(mesh, *temperature, pos);
      const ScalarOutputVal *tval = 
	dynamic_cast<const ScalarOutputVal*>(tfield.valuePtr());
      double t = tval->value();
      thermalstrain = expansiontensor*(t-tzero_);
      SymmMatrix3 thermalstress(modulus*thermalstrain);
      SymmMatrix3 strain;
      findGeometricStrain(mesh, element, pos, strain);
      double e = 0;
      for(int i=0; i<3; i++) {
	e += thermalstress(i,i)*strain(i,i);
	int j = (i+1)%3;
	e += 2*thermalstress(i,j)*strain(i,j);
      }
      *edata += -0.5*e;
    }
  }//energy output ends here
}


bool ThermalExpansion::is_symmetric(const FEMesh* mesh) const {
  Equation *forcebalance = Equation::getEquation("Force_Balance");
  // Thermal expansion makes the force-balance stiffness matrix
  // asymmetric if temperature is an active variable.
  return !(forcebalance->is_active(mesh) &&
	   temperature->is_defined(mesh) &&
	   temperature->is_active(mesh));
}

IsotropicThermalExpansion::IsotropicThermalExpansion(PyObject *registry,
						     const std::string &name, 
						     double alpha,
						     double t0)
  : ThermalExpansion(registry, name, t0),
    alpha_(alpha)
{

}


void IsotropicThermalExpansion::cross_reference(Material *mat) {
  // find out which property is the elasticity
  elasticity = dynamic_cast<Elasticity*>(mat->fetchProperty("Elasticity"));
}

void IsotropicThermalExpansion::precompute() {
  ThermalExpansion::precompute();
  expansiontensor(0,0) = expansiontensor(1,1)
    = expansiontensor(2,2) = alpha_;
}

AnisotropicThermalExpansion::AnisotropicThermalExpansion(PyObject *registry, 
							 const std::string &nm, 
							 SymmMatrix3 *alpha,
							 double t0)
  : ThermalExpansion(registry, nm, t0),
    alpha_(*alpha)
{

}
void AnisotropicThermalExpansion::cross_reference(Material *mat) {
  // find out which property is the elasticity
  elasticity = dynamic_cast<Elasticity*>(mat->fetchProperty("Elasticity"));
  orientation = dynamic_cast<OrientationProp*>
    (mat->fetchProperty("Orientation"));
}

void AnisotropicThermalExpansion::precompute()
{
  expansiontensor = alpha_.transform(orientation->eulerangle());
}
