# -*- python -*-
# $RCSfile: skeletonsegment.py,v $
# $Revision: 1.31 $
# $Author: langer $
# $Date: 2006/02/15 22:28:10 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from oof2.SWIG.common import ooferror
from oof2.engine import skeletonselectable
from oof2.common import debug
from oof2.common import object_id
from oof2.common import utils
from oof2.common import primitives
import math

# A SkeletonSegment is a connection between two nodes.  There is a
# SkeletonSegment for every adjacent pair of nodes in an element, but the
# SkeletonSegment does not belong to an element, nor does it know its
# direction.  
class SkeletonSegment(skeletonselectable.SkeletonSelectable):
    def __init__(self, nodes, index):
        ## nodes is a tuple containing two SkeletonNodes in canonical
        ## order (see utils.canonicalorder().
        skeletonselectable.SkeletonSelectable.__init__(self)
        self._nodes = nodes
        self._elements = []
        self.edges = []

        self.index = index

        ## parallel processing variables
        self._rank = -1
        self.ID = object_id.ObjectID()
    def set_rank(self, _rank):
        self.ID.set_rank(_rank)
        
    def repr_position(self):
        x = (self.nodes()[0].position().x + self.nodes()[1].position().x)/2.0
        y = (self.nodes()[0].position().y + self.nodes()[1].position().y)/2.0
        return primitives.Point(x,y)

    def length(self):
        n0 = self.nodes()[0].position()
        n1 = self.nodes()[1].position()
        return math.sqrt((n1.x-n0.x)*(n1.x-n0.x)+(n1.y-n0.y)*(n1.y-n0.y))

    def getIndex(self):
        return self.index

    def active(self, skeleton):
        return self._nodes[0].active(skeleton) or \
               self._nodes[1].active(skeleton)

    def homogeneity(self, microstructure):
        el = self._elements[0]
        nnodes = el.nnodes()
        for i in range(nnodes):
            node0 = el.nodes[i]
            node1 = el.nodes[(i+1)%nnodes]
            if self._nodes == utils.canonicalorder(node0, node1):
                return el.edgeHomogeneity(microstructure, i)

    def nodes(self):
        return self._nodes

    def addElement(self, element):
        if element in self._elements:
            debug.fmsg('*******', [n.position() for n in self._nodes])
            for e in self._elements:
                debug.fmsg([n.position() for n in e.nodes])
            debug.fmsg('adding', [n.position() for n in element.nodes])
            raise ooferror.ErrPyProgrammingError("Element already in Segment!")
        self._elements.append(element)

    def getElements(self):
        return self._elements

    def nElements(self):
        return len(self._elements)

    def removeElement(self, skeleton, element):
        self._elements.remove(element)
        if not self._elements:
            self.destroy(skeleton)
            
    def destroy(self, skeleton):
        skeleton.removeSegment(self._nodes)
        self.disconnect()
        del self._elements
        del self._nodes
        del self.edges

    # Generic node-retrieval function, common to segments and edges,
    # allowing both to be sequenced by segSequence (see below).
    def get_nodes(self):
        return self._nodes

    def get_other_node(self, node):
        for n in self._nodes:
            if n != node:
                return n
        
    # New_child is called after new nodes and elements have been
    # created -- it is assumed that the most-recent children are
    # the correct ones.  
    def new_child(self, index):
        n0 = self._nodes[0].getChildren()[-1]
        n1 = self._nodes[1].getChildren()[-1]
        new = SkeletonSegment(utils.canonicalorder(n0, n1), index)
        new._elements = [ x.getChildren()[-1] for x in self._elements ]
        return new

    def getOtherElement(self, element):
        for el in self._elements:
            if el is not element:
                return el

    def __repr__(self):
        return "SkeletonSegment(%s, %s)" % ( `self._nodes[0]`, `self._nodes[1]`)

    def get_ownership(self):
        return self._rank

    def set_ownership(self, ownr):
        self._rank = ownr
    



#######################

class SequenceError(ooferror.ErrError):
    def __init__(self, message):
        self.message = message
    def __repr__(self):
        return "SequenceError: %s" % self.message
    

# Utility function for sequencing a group of segments.  Takes as
# input an un-ordered list of segments, and an optional starting
# node, and provides as output the segments in adjacency order, plus
# the a list of nodes in the same sequence.  Should fail gracefully if
# there is a problem with the segments' adjacency.  Used by the boundary
# code.  Return value is the tuple, (segment_list, node_list)
# Calls "get_nodes" so it can also sequence SkeletonEdges.
# (In fact, it can sequence anything that has a get_nodes method...)
# The sequenced set is a loop if the node list is circular, i.e.
# node[0]==node[-1].

def segSequence(seglist, startnode=None):

    if len(seglist)==0:
        return ([], [])  # Always return the correct form, even if empty.
    
    adjacency = {}
    # Insert the segment into the adjacency dict at both of its nodes.
    for s in seglist:
        n = s.get_nodes()
        try:
            adjacency[n[0]].append(s)
        except KeyError:
            adjacency[n[0]] = [s]

        try:
            adjacency[n[1]].append(s)
        except KeyError:
            adjacency[n[1]] = [s]

    endpoints = []

    # Basic topology check -- if any node occurs in 3 or more
    # segments, there's a self-intersection, and we can't sequence.
    # This loop would also detect zero-length segment lists, but since
    # the nodes used are deduced from the passed-in segments, that
    # can't happen.
    for (n,l) in adjacency.items():
        if len(l)==1:
            endpoints.append(n)
        elif len(l)!=2:
            raise SequenceError(
                "Branch in segment set, unable to sequence.")

    # Find the user's start node, or pick our own.
    if len(endpoints)==2:  # Line case.
        # If there are two endpoints, and the passed-in start is one, OK.
        if startnode:
            if startnode == endpoints[0] or startnode == endpoints[1]:
                begin_node = startnode
            else:
                raise SequenceError(
                    "Start node not found in sequence set.")
        else:
            begin_node = endpoints[0] # Arbitrary choice.

    elif len(endpoints)==0: # Loop case.
        if startnode:
            if startnode in adjacency.keys():
                begin_node = startnode
            else:
                raise SequenceError(
                    "Start node not found in sequence loop.")
                
        else:
            begin_node = adjacency.keys()[0]  # Again, arbitrary.
    else:
        # We reach this point if the number of nodes with only one
        # corresponding segment in the passed-in set is not 0 or 2.
        # This could happen for disjoint segments that don't link up
        # properly.
        raise SequenceError(
            "Disjoint segments in segment-set, unable to sequence.")


    # Build the correctly-sequenced list now.
    seg_sequence = []
    node_sequence = [begin_node]
    while len( adjacency[node_sequence[-1]] ) > 0:
        seg = adjacency[node_sequence[-1]][0]

        seg_sequence.append(seg)
        adjacency[node_sequence[-1]].remove(seg)
        node_sequence.append(seg.get_other_node(node_sequence[-1]))

        adjacency[node_sequence[-1]].remove(seg)

    # Could remove used-up lists from the dict, then this check
    # is for any remaining entries.
    # Check completeness -- leftovers indicate an overlooked loop.
    for v in adjacency.values():
        if len(v)!=0:
            raise SequenceError(
                "Disconnected loop in segment set, unable to sequence.")

    return (seg_sequence, node_sequence)
        


# Utility function to find a path from the indicated start node to the
# indicated end node in the provided set of segments.  Returns a list
# of sequenced paths.  All of the returned paths begin at the
# indicated start point, end at the indicated end point, and do not
# repeat any segments.  They are not guaranteed to be monotonic, they
# may self-intersect, and, since all paths are returned, obviously
# they're not minimal, in general.
def segPath(start, end, segments):
    # Easy case first.
    if len(segments)==0:
        return []

    # Next easiest case -- a single segment.
    if len(segments)==1:
        n = segments[0].get_nodes()
        if (start in n) and (end in n):
            return [segments]
        else:
            return []

    # We could call segSequence at this point, which would throw an
    # exception if the segments were not a ring or a line, and thus
    # identify the path-is-obviously-unique case, but the amount of
    # effort involved is almost the same as doing the all-paths
    # construction, so we just go ahead and do that.

    # Connectivity dictionary.
    connect = {}
    for s in segments:
        n = s.get_nodes()
        try:
            connect[n[0]].append(s)
        except KeyError:
            connect[n[0]] = [s]

        try:
            connect[n[1]].append(s)
        except KeyError:
            connect[n[1]] = [s]

    # Call a helper function to recursively traverse the paths.
    return segPathFind(start, end, connect, [])


# The helper function "segPathFind" modifies the dictionary, and is
# recursive.  Return value is always a list of lists of sequenced
# segments, although this list can be empty.
def segPathFind(n1, n_target, connect, current_path):

    # Success -- if n1 is n_target, the current path is valid, return it.
    if n1==n_target:
        return [current_path]

    # If the node is not present in the dictionary, or its list
    # is of length zero, return an empty path.
    try:
        lst = connect[n1]
    except KeyError:
        return []
    else:
        if len(lst)==0:
            return []
        
    # Next-most-trivial case -- unique extension to the current path.
    if len(connect[n1])==1:
        seg = connect[n1][0]

        current_path.append(seg)
        connect[n1].remove(seg)
        # Find the other node of the segment.
        n2 = seg.get_other_node(n1)

        # Remove the other instance of the segment.
        connect[n2].remove(seg)
        # Recurse.
        return segPathFind(n2, n_target, connect, current_path)

    # Hard case -- more than one way to go at this junction.
    # Do all of them.
    seg_list = connect[n1]
    
    path_list = []
    for seg in seg_list:
        n2 = seg.get_other_node(n1)

        # Copy the dictionary.
        new_connect = {}
        for (k,v) in connect.items():
            new_connect[k]=v[:]

        # Copy the path.
        new_path = current_path[:]

        new_connect[n1].remove(seg)
        new_connect[n2].remove(seg)
        new_path.append(seg)

        path_list += segPathFind(n2, n_target, new_connect, new_path)


    return path_list



##############################

# A SkeletonEdge is a directed SkeletonSegment.  "Direction" is
# plus or minus, indicating either the same or opposite to the
# canonical ordering of the segment.
class SkeletonEdge:
    def __init__(self, segment, direction=1):
        self.segment = segment
        self.segment.edges.append(self)
        self.direction = direction

    # Useful but slow to be able to set the direction from an ordered
    # set of nodes.
    def set_direction(self, n1, n2):
        nodes = self.segment.nodes()
        if nodes[0]==n1 and nodes[1]==n2:
            self.direction = 1
        elif nodes[0]==n2 and nodes[1]==n1:
            self.direction = -1
        else:
            raise ooferror.ErrPyProgrammingError(
                "Incorrect node set in SkeletonEdge.")
    def reverse(self):
        self.direction *= -1

    # Retrieve the nodes in the correct order.  Used for drawing.
    def get_nodes(self):
        nodes = self.segment.nodes()
        if self.direction==1:
            return [nodes[0], nodes[1]]
        else:
            return [nodes[1], nodes[0]]

    def get_other_node(self, node):
        for n in self.segment.nodes():
            if node != n:
                return n
        
    # Remove yourself from your segment's edge list.
    def remove(self): 
        self.segment.edges.remove(self)

    # Function to assist with propagating boundaries.

    # This function returns a list of edges with the properties that:
    #  - The child list edges trace a path from this edge's start node
    #                      to this edge's stop node.
    #  - The child edge's segments are children of this edge's segment.
    #  - The edges are directed correctly from start to finish.
    def edgesFromSegs(self, child_segs):
        if len(child_segs)==0:
            return []
    
        # If the child segments are forked or otherwise complex,
        # return an empty list.
        try:
            (seg_list, node_list) = segSequence( child_segs )
        except SequenceError:
            return []
    
        ordered_nodes = self.get_nodes()
        #debug.fmsg('ordered_nodes=', ordered_nodes)
        #debug.fmsg('ordered_nodes[0].getChildren()=', ordered_nodes[0].getChildren())
        #debug.fmsg('node_list=', node_list)
        if ordered_nodes[0].getChildren()[0] == node_list[0]: # Forward.
            pass
        elif ordered_nodes[0].getChildren()[0] == node_list[-1]:
            seg_list.reverse()
            node_list.reverse()
        else:
            raise ooferror.ErrPyProgrammingError(
                "Malformed segment sequence -- node child not found.")
    
        child_list = []
        for (s, n)  in map(None, seg_list, node_list[:-1]):
            if s.nodes()[0]==n:
                child_list.append(SkeletonEdge(s,direction=1))
            else:
                child_list.append(SkeletonEdge(s,direction=-1))
            
        return child_list

utils.OOFdefine('SkeletonEdge', SkeletonEdge)
